#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2011-2013 Codernity (http://codernity.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import
import six

from bitdust_forks.CodernityDB3.index import (Index,
                               IndexException,
                               DocIdNotFound,
                               ElemNotFound,
                               TryReindexException,
                               IndexPreconditionsException)

import os
import hashlib
import marshal
import io
import struct
import shutil
import codecs

from bitdust_forks.CodernityDB3.storage import IU_Storage, DummyStorage

from bitdust_forks.CodernityDB3.env import cdb_environment

if cdb_environment.get('rlock_obj'):
    from bitdust_forks.CodernityDB3 import patch
    patch.patch_cache_rr(cdb_environment['rlock_obj'])

from bitdust_forks.CodernityDB3.rr_cache import cache1lvl


from bitdust_forks.CodernityDB3.misc import random_hex_32

try:
    from CodernityDB3 import __version__
except ImportError:
    from .__init__ import __version__


class IU_HashIndex(Index):
    """
    That class is for Internal Use only, if you want to use HashIndex just subclass the :py:class:`HashIndex` instead this one.

    That design is because main index logic should be always in database not in custom user indexes.
    """

    def __init__(self, db_path, name, entry_line_format='<32s{key}IIcI', hash_lim=0xfffff, storage_class=None, key_format='c'):
        """
        The index is capable to solve conflicts by `Separate chaining`
        :param db_path: database path
        :type db_path: string
        :param name: index name
        :type name: ascii string
        :param line_format: line format, `key_format` parameter value will replace `{key}` if present.
        :type line_format: string (32s{key}IIcI by default) {doc_id}{hash_key}{start}{size}{status}{next}
        :param hash_lim: maximum hash functon results (remember about birthday problem) count from 0
        :type hash_lim: integer
        :param storage_class: Storage class by default it will open standard :py:class:`CodernityDB3.storage.Storage` (if string has to be accesible by globals()[storage_class])
        :type storage_class: class name which will be instance of CodernityDB3.storage.Storage instance or None
        :param key_format: a index key format
        """
        # Fix types
        if isinstance(db_path, six.text_type):
            db_path = db_path.encode()
        if isinstance(name, six.text_type):
            name = name.encode()

        if key_format and '{key}' in entry_line_format:
            entry_line_format = entry_line_format.replace('{key}', key_format)
        super(IU_HashIndex, self).__init__(db_path, name)
        self.hash_lim = hash_lim
        if not storage_class:
            storage_class = IU_Storage
        if storage_class and not isinstance(storage_class, six.string_types):
            storage_class = storage_class.__name__
        self.storage_class = storage_class
        self.storage = None

        self.bucket_line_format = "<I"
        self.bucket_line_size = struct.calcsize(self.bucket_line_format)
        self.entry_line_format = entry_line_format
        self.entry_line_size = struct.calcsize(self.entry_line_format)

        cache = cache1lvl(100)
        self._find_key = cache(self._find_key)
        self._locate_doc_id = cache(self._locate_doc_id)
        self.bucket_struct = struct.Struct(self.bucket_line_format)
        self.entry_struct = struct.Struct(self.entry_line_format)
        self.data_start = (
            self.hash_lim + 1) * self.bucket_line_size + self._start_ind + 2

    def _fix_params(self):
        super(IU_HashIndex, self)._fix_params()
        self.bucket_line_size = struct.calcsize(self.bucket_line_format)
        self.entry_line_size = struct.calcsize(self.entry_line_format)
        self.bucket_struct = struct.Struct(self.bucket_line_format)
        self.entry_struct = struct.Struct(self.entry_line_format)
        self.data_start = (
            self.hash_lim + 1) * self.bucket_line_size + self._start_ind + 2

    def open_index(self):
        if not os.path.isfile(os.path.join(self.db_path, self.name + '_buck')):
            raise IndexException("Doesn't exists")
        self.buckets = io.open(
            os.path.join(self.db_path, self.name + "_buck"), 'r+b', buffering=0)
        self._fix_params()
        self._open_storage()

    def create_index(self):
        if os.path.isfile(os.path.join(self.db_path, self.name + '_buck')):
            raise IndexException('Already exists')
        with io.open(os.path.join(self.db_path, self.name + "_buck"), 'w+b') as f:
            props = dict(name=self.name,
                         bucket_line_format=self.bucket_line_format,
                         entry_line_format=self.entry_line_format,
                         hash_lim=self.hash_lim,
                         version=self.__version__,
                         storage_class=self.storage_class)
            f.write(marshal.dumps(props))
        self.buckets = io.open(
            os.path.join(self.db_path, self.name + "_buck"), 'r+b', buffering=0)
        self._create_storage()

    def destroy(self):
        super(IU_HashIndex, self).destroy()
        self._clear_cache()

    def _open_storage(self):
        s = globals()[self.storage_class]
        if not self.storage:
            self.storage = s(self.db_path, self.name)
        self.storage.open()

    def _create_storage(self):
        s = globals()[self.storage_class]
        if not self.storage:
            self.storage = s(self.db_path, self.name)
        self.storage.create()

    # def close_index(self):
    #     self.buckets.flush()
    #     self.buckets.close()
    #     self.storage.close()
#    @lfu_cache(100)
    def _find_key(self, key):
        """
        Find the key position

        :param key: the key to find
        """
        # print('IU_HashIndex._find_key %r' % key)
        # Fix types
        if isinstance(key, six.text_type):
            key = key.encode()

        start_position = self._calculate_position(key)
        self.buckets.seek(start_position)
        curr_data = self.buckets.read(self.bucket_line_size)
        if curr_data:
            location = self.bucket_struct.unpack(curr_data)[0]
            if not location:
                return None, None, 0, 0, 'u'
            found_at, doc_id, l_key, start, size, status, _next = self._locate_key(
                key, location)
            if status == b'd' or status == 'd':  # when first record from many is deleted
                while True:
                    found_at, doc_id, l_key, start, size, status, _next = self._locate_key(
                        key, _next)
                    if status != b'd' and status != 'd':
                        break
            return doc_id, l_key, start, size, status
        else:
            return None, None, 0, 0, 'u'

    def _find_key_many(self, key, limit=1, offset=0):
        # print('IU_HashIndex._find_key_many %r' % key)
        # Fix types
        if isinstance(key, six.text_type):
            key = key.encode()

        location = None
        start_position = self._calculate_position(key)
        self.buckets.seek(start_position)
        curr_data = self.buckets.read(self.bucket_line_size)
        if curr_data:
            location = self.bucket_struct.unpack(curr_data)[0]
        while offset:
            if not location:
                break
            try:
                found_at, doc_id, l_key, start, size, status, _next = self._locate_key(
                    key, location)
            except IndexException:
                break
            else:
                if status != b'd' and status != 'd':
                    if l_key == key:  # in case of hash function conflicts
                        offset -= 1
                location = _next
        while limit:
            if not location:
                break
            try:
                found_at, doc_id, l_key, start, size, status, _next = self._locate_key(
                    key, location)
            except IndexException:
                break
            else:
                if status != b'd' and status != 'd':
                    if l_key == key:  # in case of hash function conflicts
                        yield doc_id, start, size, status
                        limit -= 1
                location = _next

    def _calculate_position(self, key):
        # Fix types
        if isinstance(key, int):
            key = str(key)
        if isinstance(key, six.text_type):
            key = key.encode()
        pos = abs(int(hashlib.md5(key).hexdigest(), 16) & self.hash_lim) * self.bucket_line_size + self._start_ind
        # pos = abs(hash(key) & self.hash_lim) * self.bucket_line_size + self._start_ind
        # print('IU_HashIndex._calculate_position hash_lim=%r bucket_line_size=%r _start_ind=%r  %r -> %r' % (
        #     self.hash_lim, self.bucket_line_size, self._start_ind, key, pos))
        return pos

    # TODO add cache!
    def _locate_key(self, key, start):
        """
        Locate position of the key, it will iterate using `next` field in record
        until required key will be find.

        :param key: the key to locate
        :param start: position to start from
        """
        # Fix types
        if isinstance(key, six.text_type):
            key = key.encode()

        location = start
        while True:
            self.buckets.seek(location)
            data = self.buckets.read(self.entry_line_size)
            # todo, maybe partial read there...
            try:
                doc_id, l_key, start, size, status, _next = self.entry_struct.unpack(data)
                # print('IU_HashIndex._locate_key key=%r %r doc_id=%r l_key=%r start=%r size=%r status=%r _next=%r' % (key, self.name, doc_id, l_key, start, size, status, _next))
            except struct.error:
                # print('ElemNotFound 1')
                raise ElemNotFound(
                    "Not found")  # not found but might be also broken
            if l_key == key:
                break
            else:
                if not _next:
                    # print('ElemNotFound 2')
                    # not found
                    raise ElemNotFound("Not found")
                else:
                    location = _next  # go to next record
        return location, doc_id, l_key, start, size, status, _next

#    @lfu_cache(100)
    def _locate_doc_id(self, doc_id, key, start):
        """
        Locate position of the doc_id, it will iterate using `next` field in record
        until required key will be find.

        :param doc_id: the doc_id to locate
        :param key: key value
        :param start: position to start from
        """
        # Fix types
        if isinstance(doc_id, six.text_type):
            doc_id = doc_id.encode()
        if isinstance(key, six.text_type):
            key = key.encode()

        location = start
        while True:
            self.buckets.seek(location)
            data = self.buckets.read(self.entry_line_size)
            try:
                l_doc_id, l_key, start, size, status, _next = self.entry_struct.unpack(data)
            except:
                raise DocIdNotFound(
                    "Doc_id %r for %r not found" % (doc_id, key))
            if l_doc_id == doc_id and l_key == key:  # added for consistency
                break
            else:
                if not _next:
                    # not found
                    raise DocIdNotFound(
                        "Doc_id %r for %r not found" % (doc_id, key))
                else:
                    location = _next  # go to next record
        return location, doc_id, l_key, start, size, status, _next

    def _find_place(self, start):
        """
        Find a place to where put the key. It will iterate using `next` field in record, until
        empty `next` found

        :param start: position to start from
        """
        location = start
        while True:
            self.buckets.seek(location)
            data = self.buckets.read(self.entry_line_size)
            # todo, maybe partial read there...
            doc_id, l_key, start, size, status, _next = self.entry_struct.unpack(data)
            if not _next or (status == b'd' or status == 'd'):
                return self.buckets.tell() - self.entry_line_size, doc_id, l_key, start, size, status, _next
            else:
                location = _next  # go to next record

    def update(self, doc_id, key, u_start=0, u_size=0, u_status='o'):
        # print('IU_HashIndex.update %r' % key)
        # Fix types
        if isinstance(doc_id, six.text_type):
            doc_id = doc_id.encode()
        if isinstance(key, six.text_type):
            key = key.encode()
        if isinstance(u_status, six.text_type):
            u_status = u_status.encode()

        start_position = self._calculate_position(key)
        self.buckets.seek(start_position)
        curr_data = self.buckets.read(self.bucket_line_size)
        # test if it's unique or not really unique hash
        if curr_data:
            location = self.bucket_struct.unpack(curr_data)[0]
        else:
            raise ElemNotFound("Location %r not found" % doc_id)
        found_at, _doc_id, _key, start, size, status, _next = self._locate_doc_id(doc_id, key, location)
        self.buckets.seek(found_at)
        self.buckets.write(self.entry_struct.pack(doc_id,
                                                  key,
                                                  u_start,
                                                  u_size,
                                                  u_status,
                                                  _next))
        self.flush()
        self._find_key.delete(key)
        self._locate_doc_id.delete(doc_id)
        return True

    def insert(self, doc_id, key, start, size, status='o'):
        # print('IU_HashIndex.insert %r' % key)
        # Fix types
        if isinstance(doc_id, six.text_type):
            doc_id = doc_id.encode()
        if isinstance(key, six.text_type):
            key = key.encode()
        if isinstance(status, six.text_type):
            status = status.encode()

        start_position = self._calculate_position(key)
        self.buckets.seek(start_position)
        curr_data = self.buckets.read(self.bucket_line_size)

        # conflict occurs?
        if curr_data:
            location = self.bucket_struct.unpack(curr_data)[0]
        else:
            location = 0
        if location:
            # last key with that hash
            try:
                found_at, _doc_id, _key, _start, _size, _status, _next = self._locate_doc_id(doc_id, key, location)
            except DocIdNotFound:
                found_at, _doc_id, _key, _start, _size, _status, _next = self._find_place(location)
                self.buckets.seek(0, 2)
                wrote_at = self.buckets.tell()
                self.buckets.write(self.entry_struct.pack(doc_id,
                                                          key,
                                                          start,
                                                          size,
                                                          status,
                                                          _next))
#                self.flush()
                self.buckets.seek(found_at)
                self.buckets.write(self.entry_struct.pack(_doc_id,
                                                          _key,
                                                          _start,
                                                          _size,
                                                          _status,
                                                          wrote_at))
            else:
                self.buckets.seek(found_at)
                self.buckets.write(self.entry_struct.pack(doc_id,
                                                          key,
                                                          start,
                                                          size,
                                                          status,
                                                          _next))
            self.flush()
            self._locate_doc_id.delete(doc_id)
            self._find_key.delete(_key)
            # self._find_key.delete(key)
            # self._locate_key.delete(_key)
            return True
            # raise NotImplementedError
        else:
            self.buckets.seek(0, 2)
            wrote_at = self.buckets.tell()

            # check if position is bigger than all hash entries...
            if wrote_at < self.data_start:
                self.buckets.seek(self.data_start)
                wrote_at = self.buckets.tell()

            self.buckets.write(self.entry_struct.pack(doc_id,
                                                      key,
                                                      start,
                                                      size,
                                                      status,
                                                      0))
#            self.flush()
            self._find_key.delete(key)
            self.buckets.seek(start_position)
            self.buckets.write(self.bucket_struct.pack(wrote_at))
            self.flush()
            return True

    def get(self, key):
        # Fix types
        if isinstance(key, six.text_type):
            key = key.encode()
        doc_id, l_key, start, size, status = self._find_key(self.make_key(key))
        return doc_id, l_key, start, size, status

    def get_many(self, key, limit=1, offset=0):
        return self._find_key_many(self.make_key(key), limit, offset)

    def all(self, limit=-1, offset=0):
        # print('hash_index.all data_start=%r entry_line_size=%r key_format=%r' % (self.data_start, self.entry_line_size, self.key_format))
        self.buckets.seek(self.data_start)
        while offset:
            curr_data = self.buckets.read(self.entry_line_size)
            if not curr_data:
                break
            try:
                doc_id, key, start, size, status, _next = self.entry_struct.unpack(curr_data)
                # print('hash_index.all while offset doc_id=%r key=%r start=%r size=%r status=%r next=%r' % (doc_id, key, start, size, status, _next))
            except IndexException as exc:
                # print('IndexException while offset: %r' % exc)
                break
            else:
                if status != b'd' and status != 'd':
                    offset -= 1
        while limit:
            curr_data = self.buckets.read(self.entry_line_size)
            if not curr_data:
                break
            try:
                doc_id, key, start, size, status, _next = self.entry_struct.unpack(curr_data)
                # print('hash_index.all while limit doc_id=%r key=%r start=%r size=%r status=%r next=%r' % (doc_id, key, start, size, status, _next))
            except IndexException:
                # print('IndexException while limit: %r' % exc)
                break
            else:
                if status != b'd' and status != 'd':
                    yield doc_id, key, start, size, status
                    limit -= 1

    def _fix_link(self, key, pos_prev, pos_next):
        # Fix types
        if isinstance(key, six.text_type):
            key = key.encode()

        # CHECKIT why I need that hack
        if pos_prev >= self.data_start:
            self.buckets.seek(pos_prev)
            data = self.buckets.read(self.entry_line_size)
            if data:
                doc_id, l_key, start, size, status, _next = self.entry_struct.unpack(data)
                self.buckets.seek(pos_prev)
                self.buckets.write(self.entry_struct.pack(doc_id,
                                                          l_key,
                                                          start,
                                                          size,
                                                          status,
                                                          pos_next))
                self.flush()
        if pos_next:
            self.buckets.seek(pos_next)
            data = self.buckets.read(self.entry_line_size)
            if data:
                doc_id, l_key, start, size, status, _next = self.entry_struct.unpack(data)
                self.buckets.seek(pos_next)
                self.buckets.write(self.entry_struct.pack(doc_id,
                                                          l_key,
                                                          start,
                                                          size,
                                                          status,
                                                          _next))
                self.flush()
        return

    def delete(self, doc_id, key, start=0, size=0):
        # print('IU_HashIndex.delete %r' % key)
        # Fix types
        if isinstance(doc_id, six.text_type):
            doc_id = doc_id.encode()
        if isinstance(key, six.text_type):
            key = key.encode()

        start_position = self._calculate_position(key)
        self.buckets.seek(start_position)
        curr_data = self.buckets.read(self.bucket_line_size)
        if curr_data:
            location = self.bucket_struct.unpack(curr_data)[0]
        else:
            # case happens when trying to delete element with new index key in data
            # after adding new index to database without reindex
            # print("failed to delete doc_id=%r  key=%r" % (doc_id, key))
            raise TryReindexException()
        found_at, _doc_id, _key, start, size, status, _next = self._locate_doc_id(doc_id, key, location)
        self.buckets.seek(found_at)
        self.buckets.write(self.entry_struct.pack(doc_id,
                                                  key,
                                                  start,
                                                  size,
                                                  b'd',
                                                  _next))
        self.flush()
        # self._fix_link(_key, _prev, _next)
        self._find_key.delete(key)
        self._locate_doc_id.delete(doc_id)
        return True

    def compact(self, hash_lim=None):

        if not hash_lim:
            hash_lim = self.hash_lim

        compact_ind = self.__class__(
            self.db_path, self.name + '_compact', hash_lim=hash_lim)
        compact_ind.create_index()

        gen = self.all()
        while True:
            try:
                doc_id, key, start, size, status = next(gen)
            except StopIteration:
                break
            self.storage._f.seek(start)
            value = self.storage._f.read(size)
            start_ = compact_ind.storage._f.tell()
            compact_ind.storage._f.write(value)
            compact_ind.insert(doc_id, key, start_, size, status)

        compact_ind.close_index()
        original_name = self.name
        # os.unlink(os.path.join(self.db_path, self.name + "_buck"))
        self.close_index()
        shutil.move(os.path.join(compact_ind.db_path, compact_ind.
                                 name + "_buck"), os.path.join(self.db_path, self.name + "_buck"))
        shutil.move(os.path.join(compact_ind.db_path, compact_ind.
                                 name + "_stor"), os.path.join(self.db_path, self.name + "_stor"))
        # self.name = original_name
        self.open_index()  # reload...
        self.name = original_name
        self._save_params(dict(name=original_name))
        self._fix_params()
        self._clear_cache()
        return True

    def make_key(self, key):
        if isinstance(key, six.text_type):
            return codecs.encode(key, 'utf-8')
        return key

    def make_key_value(self, data):
        return b'1', data

    def _clear_cache(self):
        self._find_key.clear()
        self._locate_doc_id.clear()

    def close_index(self):
        super(IU_HashIndex, self).close_index()
        self._clear_cache()


class IU_UniqueHashIndex(IU_HashIndex):
    """
    Index for *unique* keys! Designed to be a **id** index.

    That class is for Internal Use only, if you want to use UniqueHashIndex just subclass the :py:class:`UniqueHashIndex` instead this one.

    That design is because main index logic should be always in database not in custom user indexes.
    """

    def __init__(self, db_path, name, entry_line_format="<32s8sIIcI", *args, **kwargs):
        # Fix types
        if isinstance(db_path, six.text_type):
            db_path = db_path.encode()
        if isinstance(name, six.text_type):
            name = name.encode()

        if 'key' in kwargs:
            raise IndexPreconditionsException(
                "UniqueHashIndex doesn't accept key parameter'")
        super(IU_UniqueHashIndex, self).__init__(db_path, name,
                                                 entry_line_format, *args, **kwargs)
        self.create_key = random_hex_32  # : set the function to create random key when no _id given
        # self.entry_struct=struct.Struct(entry_line_format)

#    @lfu_cache(100)
    def _find_key(self, key):
        """
        Find the key position

        :param key: the key to find
        """
        # Fix types
        # print('IU_UniqueHashIndex._find_key %r' % key)
        if isinstance(key, six.text_type):
            key = key.encode()

        start_position = self._calculate_position(key)
        self.seek = self.buckets.seek(start_position)
        curr_data = self.buckets.read(self.bucket_line_size)
        if curr_data:
            location = self.bucket_struct.unpack(curr_data)[0]
            found_at, l_key, rev, start, size, status, _next = self._locate_key(
                key, location)

            # Fix types
            if isinstance(l_key, bytes):
                l_key = l_key.decode()
            if isinstance(rev, bytes):
                rev = rev.decode()
            if isinstance(status, bytes):
                status = status.decode()

            return l_key, rev, start, size, status
        else:
            return None, None, 0, 0, 'u'

    def _find_key_many(self, *args, **kwargs):
        raise NotImplementedError()

    def _find_place(self, start, key):
        """
        Find a place to where put the key. It will iterate using `next` field in record, until
        empty `next` found

        :param start: position to start from
        """
        # Fix types
        if isinstance(key, six.text_type):
            key = key.encode()

        location = start
        while True:
            self.buckets.seek(location)
            data = self.buckets.read(self.entry_line_size)
            # todo, maybe partial read there...
            l_key, rev, start, size, status, _next = self.entry_struct.unpack(
                data)
            if l_key == key:
                raise IndexException("The %r key already exists" % key)
            if not _next or (status == b'd' or status == 'd'):
                return self.buckets.tell() - self.entry_line_size, l_key, rev, start, size, status, _next
            else:
                location = _next  # go to next record

    # @lfu_cache(100)
    def _locate_key(self, key, start):
        """
        Locate position of the key, it will iterate using `next` field in record
        until required key will be find.

        :param key: the key to locate
        :param start: position to start from
        """
        # Fix types
        if isinstance(key, six.text_type):
            key = key.encode()

        location = start
        while True:
            self.buckets.seek(location)
            data = self.buckets.read(self.entry_line_size)
            # todo, maybe partial read there...
            try:
                l_key, rev, start, size, status, _next = self.entry_struct.unpack(data)
                # print('IU_UniqueHashIndex._locate_key %r key=%r start=%r location=%r l_key=%r rev=%r start=%r size=%r status=%r _next=%r' % (
                #     self.name, key, start, location, l_key, rev, start, size, status, _next))
            except struct.error:
                # print("Location %r not found 1" % key)
                raise ElemNotFound("Location %r not found" % key)
            if l_key == key:
                break
            else:
                if not _next:
                    # not found
                    # print("Location %r not found 2" % key)
                    raise ElemNotFound("Location %r not found" % key)
                else:
                    location = _next  # go to next record
        return self.buckets.tell() - self.entry_line_size, l_key, rev, start, size, status, _next

    def update(self, key, rev, u_start=0, u_size=0, u_status='o'):
        # print('IU_UniqueHashIndex.update %r' % key)
        # Fix types
        if isinstance(key, six.text_type):
            key = key.encode()
        if isinstance(rev, six.text_type):
            rev = rev.encode()
        if isinstance(u_status, six.text_type):
            u_status = u_status.encode()

        start_position = self._calculate_position(key)
        self.buckets.seek(start_position)
        curr_data = self.buckets.read(self.bucket_line_size)
        # test if it's unique or not really unique hash

        if curr_data:
            location = self.bucket_struct.unpack(curr_data)[0]
        else:
            raise ElemNotFound("Location %r not found" % key)
        found_at, _key, _rev, start, size, status, _next = self._locate_key(
            key, location)
        if u_start == 0:
            u_start = start
        if u_size == 0:
            u_size = size
        self.buckets.seek(found_at)
        self.buckets.write(self.entry_struct.pack(key,
                                                  rev,
                                                  u_start,
                                                  u_size,
                                                  u_status,
                                                  _next))
        self.flush()
        self._find_key.delete(key)
        return True

    def insert(self, key, rev, start, size, status='o'):
        # print('IU_UniqueHashIndex.insert %r' % key)
        # Fix types
        if isinstance(key, six.text_type):
            key = key.encode()
        if isinstance(rev, six.text_type):
            rev = rev.encode()
        if isinstance(status, six.text_type):
            status = status.encode()

        start_position = self._calculate_position(key)
        self.buckets.seek(start_position)
        curr_data = self.buckets.read(self.bucket_line_size)

        # conflict occurs?
        if curr_data:
            location = self.bucket_struct.unpack(curr_data)[0]
        else:
            location = 0
        if location:
            # last key with that hash
            found_at, _key, _rev, _start, _size, _status, _next = self._find_place(
                location, key)
            self.buckets.seek(0, 2)
            wrote_at = self.buckets.tell()

            # check if position is bigger than all hash entries...
            if wrote_at < self.data_start:
                self.buckets.seek(self.data_start)
                wrote_at = self.buckets.tell()

            self.buckets.write(self.entry_struct.pack(key,
                                                      rev,
                                                      start,
                                                      size,
                                                      status,
                                                      _next))

#            self.flush()
            self.buckets.seek(found_at)
            self.buckets.write(self.entry_struct.pack(_key,
                                                      _rev,
                                                      _start,
                                                      _size,
                                                      _status,
                                                      wrote_at))
            self.flush()
            self._find_key.delete(_key)
            # self._locate_key.delete(_key)
            return True
            # raise NotImplementedError
        else:
            self.buckets.seek(0, 2)
            wrote_at = self.buckets.tell()

            # check if position is bigger than all hash entries...
            if wrote_at < self.data_start:
                self.buckets.seek(self.data_start)
                wrote_at = self.buckets.tell()

            self.buckets.write(self.entry_struct.pack(key,
                                                      rev,
                                                      start,
                                                      size,
                                                      status,
                                                      0))

            self.buckets.seek(start_position)
            self.buckets.write(self.bucket_struct.pack(wrote_at))
            self.flush()
            self._find_key.delete(key)
            return True

    def all(self, limit=-1, offset=0):
        self.buckets.seek(self.data_start)
        while offset:
            curr_data = self.buckets.read(self.entry_line_size)
            if not curr_data:
                break
            try:
                doc_id, rev, start, size, status, next = self.entry_struct.unpack(curr_data)
                # print('IU_UniqueHashIndex.all while offset %r doc_id=%r rev=%r start=%r size=%r status=%r next=%r' % (self.name, doc_id, rev, start, size, status, next))
            except IndexException:
                break
            else:
                if status != b'd' and status != 'd':
                    offset -= 1

        while limit:
            curr_data = self.buckets.read(self.entry_line_size)
            if not curr_data:
                break
            try:
                doc_id, rev, start, size, status, next = self.entry_struct.unpack(curr_data)
                # print('IU_UniqueHashIndex.all while limit %r doc_id=%r rev=%r start=%r size=%r status=%r next=%r' % (self.name, doc_id, rev, start, size, status, next))
            except IndexException:
                break
            else:
                if status != b'd' and status != 'd':
                    if isinstance(rev, six.binary_type):
                        rev = rev.decode()
                    if isinstance(doc_id, six.binary_type):
                        doc_id = doc_id.decode()
                    yield doc_id, rev, start, size, status
                    limit -= 1

    def get_many(self, *args, **kwargs):
        raise NotImplementedError()

    def delete(self, key, start=0, size=0):
        # Fix types
        if isinstance(key, six.text_type):
            key = key.encode()
        self.update(key, '00000000', start, size, b'd')

    def make_key_value(self, data):
        _id = data['_id']
        try:
            _id = data['_id']
            if isinstance(_id, six.text_type):
                _id = _id.encode()
        except:
            raise IndexPreconditionsException(
                "_id must be valid string/bytes object")
        if not isinstance(_id, bytes) or len(_id) != 32:
            raise IndexPreconditionsException("Invalid _id type or lenght")
        del data['_id']
        del data['_rev']
        return _id, data

    def destroy(self):
        Index.destroy(self)
        self._clear_cache()

    def _clear_cache(self):
        self._find_key.clear()

    def insert_with_storage(self, _id, _rev, value):
        # Fix types
        if isinstance(_id, six.text_type):
            _id = _id.encode()
        if isinstance(_rev, six.text_type):
            _rev = _rev.encode()
        if value:
            start, size = self.storage.insert(value)
        else:
            start = 1
            size = 0
        return self.insert(_id, _rev, start, size)

    def update_with_storage(self, _id, _rev, value):
        # Fix types
        if isinstance(_id, six.text_type):
            _id = _id.encode()
        if isinstance(_rev, six.text_type):
            _rev = _rev.encode()

        if value:
            start, size = self.storage.insert(value)
        else:
            start = 1
            size = 0
        return self.update(_id, _rev, start, size)


class DummyHashIndex(IU_HashIndex):
    def __init__(self, db_path, name, entry_line_format="<32s4sIIcI", *args, **kwargs):
        super(DummyHashIndex, self).__init__(db_path, name,
                                             entry_line_format, *args, **kwargs)
        self.create_key = random_hex_32  # : set the function to create random key when no _id given
        # self.entry_struct=struct.Struct(entry_line_format)

    def update(self, *args, **kwargs):
        return True

    def insert(self, *args, **kwargs):
        return True

    def all(self, *args, **kwargs):
        raise StopIteration

    def get(self, *args, **kwargs):
        raise ElemNotFound

    def get_many(self, *args, **kwargs):
        raise StopIteration

    def delete(self, *args, **kwargs):
        pass

    def make_key_value(self, data):
        return '1', {'_': 1}

    def destroy(self):
        pass

    def _clear_cache(self):
        pass

    def _open_storage(self):
        if not self.storage:
            self.storage = DummyStorage()
        self.storage.open()

    def _create_storage(self):
        if not self.storage:
            self.storage = DummyStorage()
        self.storage.create()


class IU_MultiHashIndex(IU_HashIndex):
    """
    Class that allows to index more than one key per database record.

    It operates very well on GET/INSERT. It's not optimized for
    UPDATE operations (will always readd everything)
    """

    def __init__(self, *args, **kwargs):
        super(IU_MultiHashIndex, self).__init__(*args, **kwargs)

    def insert(self, doc_id, key, start, size, status='o'):
        if isinstance(key, (list, tuple)):
            key = set(key)
        elif not isinstance(key, set):
            key = set([key])
        ins = super(IU_MultiHashIndex, self).insert
        for curr_key in key:
            ins(doc_id, curr_key, start, size, status)
        return True

    def update(self, doc_id, key, u_start, u_size, u_status='o'):
        if isinstance(key, (list, tuple)):
            key = set(key)
        elif not isinstance(key, set):
            key = set([key])
        upd = super(IU_MultiHashIndex, self).update
        for curr_key in key:
            upd(doc_id, curr_key, u_start, u_size, u_status)

    def delete(self, doc_id, key, start=0, size=0):
        if isinstance(key, (list, tuple)):
            key = set(key)
        elif not isinstance(key, set):
            key = set([key])
        delete = super(IU_MultiHashIndex, self).delete
        for curr_key in key:
            delete(doc_id, curr_key, start, size)

    def get(self, key):
        return super(IU_MultiHashIndex, self).get(key)

    def make_key_value(self, data):
        raise NotImplementedError()


# classes for public use, done in this way because of
# generation static files with indexes (_index directory)


class HashIndex(IU_HashIndex):
    """
    That class is designed to be used in custom indexes.
    """
    pass


class UniqueHashIndex(IU_UniqueHashIndex):
    """
    That class is designed to be used in custom indexes. It's designed to be **id** index.
    """
    pass


class MultiHashIndex(IU_MultiHashIndex):
    """
    That class is designed to be used in custom indexes.
    """



import threading
import time


class HyperlaneManager(threading.Thread):
    def __init__(self, app_log):
        threading.Thread.__init__(self)
        self.app_log = app_log

    def run(self):
        self.hyperlane_manager()

    def hyperlane_manager(self):
        self.app_log.warning('Hyperlane manager initiated')
        while True:
            self.app_log.warning('Hyperlane manager running')
            time.sleep(5)


if __name__ == '__main__':
    import options
    import log

    config = options.Get()
    config.read()
    app_log = log.log('hyperlane.log', 'WARNING', True)

    hyperlane_manager = HyperlaneManager(app_log)
    hyperlane_manager.start()

    app_log.warning('Loop started')



#!/usr/bin/python
# net_misc.py
#
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (net_misc.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: net_misc.

Some network routines
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
import six
import six.moves.urllib.request, six.moves.urllib.parse, six.moves.urllib.error  # @UnresolvedImport
import six.moves.urllib.request, six.moves.urllib.error, six.moves.urllib.parse  # @UnresolvedImport
import six.moves.urllib.parse  # @UnresolvedImport
from io import open

#------------------------------------------------------------------------------

import os
import re
import sys
import socket
import random
import platform
import mimetypes
import subprocess

#------------------------------------------------------------------------------

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in net_misc.py')

from twisted.python import log as twisted_log  # @UnresolvedImport
from twisted.python.failure import Failure  # @UnresolvedImport
from twisted.internet import protocol  # @UnresolvedImport
from twisted.internet.defer import Deferred, DeferredList, succeed, fail, CancelledError  # @UnresolvedImport
from twisted.web import iweb, client, http_headers  # @UnresolvedImport
from twisted.web.client import Agent, _ReadBodyProtocol  # @UnresolvedImport
from twisted.web.http_headers import Headers  # @UnresolvedImport

from zope.interface import implementer

#------------------------------------------------------------------------------

from bitdust.lib import strng

#------------------------------------------------------------------------------

_ConnectionDoneCallbackFunc = None
_ConnectionFailedCallbackFunc = None

#------------------------------------------------------------------------------

_UserAgentString = 'BitDust-http-agent'
_ProxySettings = {'host': '', 'port': '', 'ssl': 'False', 'username': '', 'password': ''}

#------------------------------------------------------------------------------


def init():
    pass


def shutdown():
    pass


#------------------------------------------------------------------------------


def SetConnectionDoneCallbackFunc(f):
    """
    Here is a place to set a callback to catch events for ````successful````
    network transfers.
    """
    global _ConnectionDoneCallbackFunc
    _ConnectionDoneCallbackFunc = f


def SetConnectionFailedCallbackFunc(f):
    """
    Set a callback to catch events for ````failed```` network transfers or
    connections.

    Later, BitDust code will compare both counters to decide that
    connection to Internet is gone.
    """
    global _ConnectionFailedCallbackFunc
    _ConnectionFailedCallbackFunc = f


def ConnectionDone(param=None, proto=None, info=None):
    """
    This method is called from different places to inform of ````successful````
    network transfers, connections, requests.
    """
    global _ConnectionDoneCallbackFunc
    if _ConnectionDoneCallbackFunc is not None:
        _ConnectionDoneCallbackFunc(proto, info, param)
    return param


def ConnectionFailed(param=None, proto=None, info=None):
    """
    This method is called to inform of ````failed```` network results.
    """
    global _ConnectionFailedCallbackFunc
    if _ConnectionFailedCallbackFunc is not None:
        _ConnectionFailedCallbackFunc(proto, info, param)
    if param and isinstance(param, Failure) and param.type == CancelledError:
        return None
    return param


#------------------------------------------------------------------------------


def normalize_address(host_port, default_port=80):
    """
    Input argument `host` can be string: "123.45.67.89:8080" or tuple: (b"123.45.67.89", 8080)
    Method always return tuple and make sure host is a binary string and port is an integer.
    """
    if not host_port:
        return host_port
    if isinstance(host_port, six.binary_type):
        host_port = host_port.decode('utf-8')
    if isinstance(host_port, six.string_types):
        host_port = host_port.split(':')
        host_port = (
            host_port[0],
            int(host_port[1]),
        )
    if not host_port[1]:
        host_port = (
            host_port[0],
            default_port,
        )
    if isinstance(host_port[0], six.binary_type):
        host_port = (
            host_port[0].decode('utf-8'),
            int(host_port[1]),
        )
    if isinstance(host_port[0], six.text_type):
        host_port = (
            host_port[0].encode('utf-8'),
            int(host_port[1]),
        )
    return host_port


def pack_address(host_port, proto=None, default_port=80):
    """
    Same as `normalize_address()`, but always return address as byte string: b"123.45.67.89:8080"
    """
    if not host_port:
        if proto:
            return strng.to_bin(proto) + b'://'
        return host_port
    norm = normalize_address(host_port, default_port=default_port)
    if proto:
        return strng.to_bin(proto) + b'://' + norm[0] + b':' + str(norm[1]).encode()
    return norm[0] + b':' + str(norm[1]).encode()


def pack_address_text(host_port, proto=None, default_port=80):
    """
    Same as `pack_address()`, but returns text string or None.
    """
    if not host_port:
        return None
    return strng.to_text(pack_address(host_port, proto=proto, default_port=default_port))


#------------------------------------------------------------------------------


def parse_url(url, defaultPort=None):
    """
    Split the given URL into the scheme, host, port, and path.
    """
    url = strng.to_text(url)
    url = url.strip()
    parsed = six.moves.urllib.parse.urlparse(url)
    scheme = parsed[0]
    path = six.moves.urllib.parse.urlunparse(('', '') + parsed[2:])
    if defaultPort is None:
        if scheme == 'https':
            defaultPort = 443
        else:
            defaultPort = 80
    host, port = parsed[1], defaultPort
    if ':' in host:
        host, port = host.rsplit(':', 1)
        try:
            port = int(port)
        except ValueError:
            port = defaultPort
    if not path:
        path = '/'
    return scheme, host, port, path


def parse_credentials(host):
    """
    Test host name (network location) for credentials and split by parts: host,
    username, password.
    """
    if not host.count('@'):
        return host, '', ''
    credentials, host = host.rsplit('@', 1)
    if not credentials.count(':'):
        return host, credentials, ''
    username, password = credentials.split(':', 1)
    return host, username, password


#------------------------------------------------------------------------------


def detect_proxy_settings():
    """
    Do some work and return dictionary with Proxy server settings for that
    machine.
    """
    d = {'host': '', 'port': '', 'username': '', 'password': '', 'ssl': 'False'}
    return d

    # TODO: to be fixed later
    # import urllib2
    # httpproxy = urllib2.getproxies().get('http', None)
    # httpsproxy = urllib2.getproxies().get('https', None)
    # if httpproxy is not None:
    #     try:
    #         scheme, host, port, path = parse_url(httpproxy)
    #         host, username, password = parse_credentials(host)
    #     except:
    #         return d
    #     d['ssl'] = 'False'
    #     d['host'] = host
    #     d['port'] = port
    #     d['username'] = username
    #     d['password'] = password
    # if httpsproxy is not None:
    #     try:
    #         scheme, host, port, path = parse_url(httpsproxy)
    #         host, username, password = parse_credentials(host)
    #     except:
    #         return d
    #     d['ssl'] = 'True'
    #     d['host'] = host
    #     d['port'] = port
    #     d['username'] = username
    #     d['password'] = password
    # return d


def set_proxy_settings(settings_dict):
    """
    Remember proxy settings.
    """
    global _ProxySettings
    _ProxySettings = settings_dict


def get_proxy_settings():
    """
    Get current proxy settings.
    """
    global _ProxySettings
    return _ProxySettings


def get_proxy_host():
    """
    Get current proxy host.
    """
    global _ProxySettings
    return _ProxySettings.get('host', '')


def get_proxy_port():
    """
    Get current proxy port number.
    """
    global _ProxySettings
    try:
        return int(_ProxySettings.get('port', '8080'))
    except:
        return 8080


def get_proxy_username():
    """
    Get current proxy username.
    """
    global _ProxySettings
    return _ProxySettings.get('username', '')


def get_proxy_password():
    """
    Get current proxy password.
    """
    global _ProxySettings
    return _ProxySettings.get('password', '')


def get_proxy_ssl():
    """
    Is this a secure proxy?
    """
    global _ProxySettings
    return _ProxySettings.get('ssl', '')


def proxy_is_on():
    """
    In most cases people do not use any proxy servers.

    This is to check if user is using a proxy and we have the settings.
    """
    return get_proxy_host() != ''


#-------------------------------------------------------------------------------


class custom_ReadBodyProtocol(_ReadBodyProtocol):
    def connectionLost(self, reason):
        if self.deferred.called:
            return
        super(custom_ReadBodyProtocol, self).connectionLost(reason)


def readBody(response):
    def cancel(deferred):
        abort = getAbort()
        if abort is not None:
            abort()

    d = Deferred(cancel)
    d.addErrback(twisted_log.err)
    protocol = custom_ReadBodyProtocol(response.code, response.phrase, d)

    def getAbort():
        return getattr(protocol.transport, 'abortConnection', None)

    response.deliverBody(protocol)
    return d


def readBodyFailed(result):
    if result.type == CancelledError:
        return None
    return result


def readResponse(response, timeout):
    #     print('Response version:', response.version)
    #     print('Response code:', response.code)
    #     print('Response phrase:', response.phrase)
    #     print('Response headers:')
    #     print(list(response.headers.getAllRawHeaders()))
    if response.code != 200:
        return fail(Exception('Bad response from the server: [%d] %s' % (
            response.code,
            response.phrase.strip(),
        ), ))
    d = readBody(response)
    d.addTimeout(timeout=timeout, clock=reactor)
    d.addErrback(readBodyFailed)
    return d


def getPageTwisted(url, timeout=15, method=b'GET'):
    """
    A smart way to download pages from HTTP hosts.
    """
    #     def getPageTwistedTimeout(_d):
    #         _d.cancel()
    #     def getPageTwistedTimeoutDisconnect(_tcpcall):
    #         _tcpcall.disconnect()
    #     def getPageTwistedCancelTimeout(x, _t):
    #         if _t.active():
    #             _t.cancel()
    #         return x
    global _UserAgentString

    url = strng.to_bin(url)

    #     if proxy_is_on():
    #         factory = ProxyClientFactory(url, agent=_UserAgentString, timeout=timeout)
    #         tcp_call = reactor.connectTCP(get_proxy_host(), get_proxy_port(), factory)
    # #         if timeout:
    # #             timeout_call = reactor.callLater(timeout, getPageTwistedTimeoutDisconnect, tcp_call)
    # #             factory.deferred.addBoth(getPageTwistedCancelTimeout, timeout_call)
    #         factory.deferred.addCallback(ConnectionDone, 'http', 'getPageTwisted proxy %s' % (url))
    #         factory.deferred.addErrback(ConnectionFailed, 'http', 'getPageTwisted proxy %s' % (url))
    #         return factory.deferred

    agent = Agent(reactor, connectTimeout=timeout)

    d = agent.request(
        method=method,
        uri=url,
        headers=Headers({b'User-Agent': [_UserAgentString]}),
    )
    #     d = getPage(url, agent=_UserAgentString, timeout=timeout)
    #     if timeout:
    #         timeout_call = reactor.callLater(timeout, getPageTwistedTimeout, d)
    #         d.addBoth(getPageTwistedCancelTimeout, timeout_call)
    d.addCallback(readResponse, timeout)
    # d.addTimeout(timeout=timeout, clock=reactor)
    d.addCallback(ConnectionDone, 'http', 'getPageTwisted %r' % url)
    d.addErrback(ConnectionFailed, 'http', 'getPageTwisted %r' % url)
    return d


#------------------------------------------------------------------------------


def IpIsLocal(ip):
    """
    A set of "classic" patterns for local networks.
    """
    if not ip:
        return True
    if ip == '0.0.0.0':
        return True
    if ip.startswith('192.168.'):
        return True
    if ip.startswith('10.'):
        return True
    if ip.startswith('127.'):
        return True
    if ip.startswith('172.'):
        try:
            secondByte = int(ip.split('.')[1])
        except:
            raise Exception('wrong ip address ' + str(ip))
            return True
        if secondByte >= 16 and secondByte <= 31:
            return True
    return False


def getLocalIp():
    """
    A stack of methods to get the local IP of that machine.

    Had this in p2p/stun.py.
    http://ubuntuforums.org/showthread.php?t=1215042
    1. Use the gethostname method
    2. Use outside connection
    3. Use OS specific command
    4. Return 127.0.0.1 in unknown situation
    """
    # 1: Use the gethostname method

    try:
        ipaddr = socket.gethostbyname(socket.gethostname())
        if not (ipaddr.startswith('127')):
            #print('Can use Method 1: ' + ipaddr)
            return ipaddr
    except:
        pass

    # 2: Use outside connection
    '''
    Source:
    http://commandline.org.uk/python/how-to-find-out-ip-address-in-python/
    '''

    ipaddr = ''
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect(('google.com', 0))
        ipaddr = s.getsockname()[0]
        #print('Can used Method 2: ' + ipaddr)
        return ipaddr
    except:
        pass

    # 3: Use OS specific command
    ipaddr = ''
    os_str = platform.system().upper()

    if os_str == 'LINUX':

        # Linux:
        arg = '`which ip` route list'
        p = subprocess.Popen(arg, shell=True, stdout=subprocess.PIPE)
        data = p.communicate()
        sdata = data[0].split()
        ipaddr = sdata[sdata.index('src') + 1]
        #netdev = sdata[ sdata.index('dev')+1 ]
        #print('Can used Method 3: ' + ipaddr)
        return ipaddr

    elif os_str == 'WINDOWS':

        # Windows:
        arg = 'route print 0.0.0.0'
        p = subprocess.Popen(arg, shell=True, stdout=subprocess.PIPE)
        data = p.communicate()
        strdata = data[0].decode()
        sdata = strdata.split()

        while len(sdata) > 0:
            if sdata.pop(0) == 'Netmask':
                if sdata[0] == 'Gateway' and sdata[1] == 'Interface':
                    ipaddr = sdata[6]
                    break
        #print('Can used Method 4: ' + ipaddr)
        return ipaddr

    return '127.0.0.1'  # uh oh, we're in trouble, but don't want to return none


#-------------------------------------------------------------------------------


def TestInternetConnectionOld(remote_host='www.google.com'):  # 74.125.113.99
    """
    Ancient method to check Internet connection.
    """
    try:
        (family, socktype, proto, garbage, address) = socket.getaddrinfo(remote_host, 'http')[0]
    except Exception as e:
        return False

    s = socket.socket(family, socktype, proto)

    try:
        result = s.connect(address)
    except Exception as e:
        return False

    return result is None or result == 0


#------------------------------------------------------------------------------


def TestInternetConnectionOld2(remote_hosts=None, timeout=15):
    """
    A little bit more smart method to check Internet connection.
    """
    if remote_hosts is None:
        remote_hosts = []
        remote_hosts.append('www.google.com')
        remote_hosts.append('www.facebook.com')
        remote_hosts.append('www.youtube.com')
        # remote_hosts.append('www.yahoo.com')
        # remote_hosts.append('www.baidu.com')

    def _response(src, result):
        # print '_response', err, hosts, index
        result.callback(None)

    def _fail(err, hosts, index, result):
        # print 'fail', hosts, index
        reactor.callLater(0, _call, hosts, index + 1, result)  # @UndefinedVariable

    def _call(hosts, index, result):
        # print 'call' , hosts, index, result
        if index >= len(hosts):
            result.errback(None)
            return
        # print '    ', hosts[index]
        d = getPageTwisted(hosts[index])
        d.addCallback(_response, result)
        d.addErrback(_fail, hosts, index, result)

    result = Deferred()
    reactor.callLater(0, _call, remote_hosts, 0, result)  # @UndefinedVariable
    return result


#------------------------------------------------------------------------------


def TestInternetConnection(remote_hosts=None, timeout=15):
    if remote_hosts is None:
        remote_hosts = []
        from bitdust.userid import known_servers
        for host, ports in known_servers.by_host().items():
            remote_hosts.append('http://%s:%d' % (
                host,
                ports[0],
            ))
    random.shuffle(remote_hosts)
    dl = []
    for host in remote_hosts[:5]:
        dl.append(getPageTwisted(host, timeout=timeout))
    return DeferredList(dl, fireOnOneCallback=True, fireOnOneErrback=False, consumeErrors=True)


#------------------------------------------------------------------------------


def uploadHTTP(url, files, data, progress=None, receiverDeferred=None):
    """
    A smart way to upload a file over HTTP POST method.

    Use ``MultiPartProducer`` and ``StringReceiver`` classes.
    Great Thanks to Mariano!

    http://marianoiglesias.com.ar/python/file-uploading-with-multi-part-encoding-using-twisted/
    """
    class StringReceiver(protocol.Protocol):
        buffer = ''

        def __init__(self, deferred=None):
            self._deferred = deferred

        def dataReceived(self, data):
            self.buffer += data

        def connectionLost(self, reason):
            if self._deferred and reason.check(client.ResponseDone):
                self._deferred.callback(self.buffer)
            else:
                self._deferred.errback(Exception(self.buffer))

    @implementer(iweb.IBodyProducer)
    class MultiPartProducer:

        CHUNK_SIZE = 2**8

        def __init__(self, files={}, data={}, callback=None, deferred=None):
            self._files = files
            self._file_lengths = {}
            self._data = data
            self._callback = callback
            self._deferred = deferred
            self.boundary = self._boundary()
            self.length = self._length()

        def startProducing(self, consumer):
            self._consumer = consumer
            self._current_deferred = Deferred()
            self._sent = 0
            self._paused = False
            if not hasattr(self, '_chunk_headers'):
                self._build_chunk_headers()
            if self._data:
                block = ''
                for field in self._data:
                    block += self._chunk_headers[field]
                    block += self._data[field]
                    block += '\r\n'
                self._send_to_consumer(block)
            if self._files:
                self._files_iterator = six.iterkeys(self._files)
                self._files_sent = 0
                self._files_length = len(self._files)
                self._current_file_path = None
                self._current_file_handle = None
                self._current_file_length = None
                self._current_file_sent = 0
                result = self._produce()
                if result:
                    return result
            else:
                return succeed(None)
            return self._current_deferred

        def resumeProducing(self):
            self._paused = False
            result = self._produce()
            if result:
                return result

        def pauseProducing(self):
            self._paused = True

        def stopProducing(self):
            self._finish(True)
            if self._deferred and self._sent < self.length:
                self._deferred.errback(Exception('Consumer asked to stop production of request body (%d sent out of %d)' % (self._sent, self.length)))

        def _produce(self):
            if self._paused:
                return
            done = False
            while not done and not self._paused:
                if not self._current_file_handle:
                    field = next(self._files_iterator)
                    self._current_file_path = self._files[field]
                    self._current_file_sent = 0
                    self._current_file_length = self._file_lengths[field]
                    self._current_file_handle = open(self._current_file_path, 'rb')
                    self._send_to_consumer(self._chunk_headers[field])
                chunk = self._current_file_handle.read(self.CHUNK_SIZE)
                if chunk:
                    self._send_to_consumer(chunk)
                    self._current_file_sent += len(chunk)
                if not chunk or self._current_file_sent == self._current_file_length:
                    self._send_to_consumer('\r\n')
                    self._current_file_handle.close()
                    self._current_file_handle = None
                    self._current_file_sent = 0
                    self._current_file_path = None
                    self._files_sent += 1
                if self._files_sent == self._files_length:
                    done = True
            if done:
                self._send_to_consumer('--%s--\r\n' % self.boundary)
                self._finish()
                return succeed(None)

        def _finish(self, forced=False):
            if hasattr(self, '_current_file_handle') and self._current_file_handle:
                self._current_file_handle.close()
            if self._current_deferred:
                self._current_deferred.callback(self._sent)
                self._current_deferred = None
            if not forced and self._deferred:
                self._deferred.callback(self._sent)

        def _send_to_consumer(self, block):
            self._consumer.write(block)
            self._sent += len(block)
            if self._callback:
                self._callback(self._sent, self.length)

        def _length(self):
            self._build_chunk_headers()
            length = 0
            if self._data:
                for field in self._data:
                    length += len(self._chunk_headers[field])
                    length += len(self._data[field])
                    length += 2
            if self._files:
                for field in self._files:
                    length += len(self._chunk_headers[field])
                    length += self._file_size(field)
                    length += 2
            length += len(self.boundary)
            length += 6
            return length

        def _build_chunk_headers(self):
            if hasattr(self, '_chunk_headers') and self._chunk_headers:
                return
            self._chunk_headers = {}
            for field in self._files:
                self._chunk_headers[field] = self._headers(field, True)
            for field in self._data:
                self._chunk_headers[field] = self._headers(field)

        def _headers(self, name, is_file=False):
            value = self._files[name] if is_file else self._data[name]
            _boundary = self.boundary.encode('utf-8') if isinstance(self.boundary, six.text_type) else six.moves.urllib.parse.quote_plus(self.boundary)
            headers = ['--%s' % _boundary]
            if is_file:
                disposition = 'form-data; name="%s"; filename="%s"' % (name, os.path.basename(value))
            else:
                disposition = 'form-data; name="%s"' % name
            headers.append('Content-Disposition: %s' % disposition)
            if is_file:
                file_type = self._file_type(name)
            else:
                file_type = 'text/plain; charset=utf-8'
            headers.append('Content-Type: %s' % file_type)
            if is_file:
                headers.append('Content-Length: %i' % self._file_size(name))
            else:
                headers.append('Content-Length: %i' % len(value))
            headers.append('')
            headers.append('')
            return '\r\n'.join(headers)

        def _boundary(self):
            boundary = None
            try:
                import uuid
                boundary = uuid.uuid4().hex
            except ImportError:
                import random
                import sha
                bits = random.getrandbits(160)
                boundary = sha.new(str(bits).encode()).hexdigest()
            return boundary

        def _file_type(self, field):
            typ = mimetypes.guess_type(self._files[field])[0]
            return typ.encode('utf-8') if isinstance(typ, six.text_type) else str(typ)

        def _file_size(self, field):
            size = 0
            try:
                handle = open(self._files[field], 'r')
                size = os.fstat(handle.fileno()).st_size
                handle.close()
            except:
                size = 0
            self._file_lengths[field] = size
            return self._file_lengths[field]

    # producerDeferred = Deferred()
    receiverDeferred = Deferred()

    myProducer = MultiPartProducer(files, data, progress)  # , producerDeferred)
    myReceiver = StringReceiver(receiverDeferred)

    headers = http_headers.Headers()
    headers.addRawHeader('Content-Type', 'multipart/form-data; boundary=%s' % myProducer.boundary)

    agent = client.Agent(reactor)
    request = agent.request(b'POST', url, headers, myProducer)
    request.addCallback(lambda response: response.deliverBody(myReceiver))
    return request


#------------------------------------------------------------------------------


@implementer(iweb.IBodyProducer)
class BytesProducer(object):
    def __init__(self, body):
        self.body = body
        self.length = len(body)

    def startProducing(self, consumer):
        consumer.write(self.body)
        return succeed(None)

    def pauseProducing(self):
        pass

    def stopProducing(self):
        pass


def http_post_data(url, data, connectTimeout=15):
    agent = Agent(reactor, connectTimeout=connectTimeout)
    body = BytesProducer(data)
    requested = agent.request(
        b'POST',
        url,
        Headers({
            'User-Agent': ['BitDust HTTP client'],
        }),
        body,
    )
    return requested


#------------------------------------------------------------------------------


def getIfconfig(iface='en0'):
    try:
        result = subprocess.check_output('ifconfig %s | grep -w inet' % (iface), shell=True, stderr=subprocess.STDOUT)
    except:
        return None
    ip = ''
    if result:
        strs = result.split(b'\n')
        for line in strs:
            # remove \t, space...
            line = line.strip()
            if line.startswith(b'inet '):
                a = line.find(b' ')
                ipStart = a + 1
                ipEnd = line.find(b' ', ipStart)
                if a != -1 and ipEnd != -1:
                    ip = line[ipStart:ipEnd]
                    break
    return ip


#------------------------------------------------------------------------------


def getNetworkInterfaces():
    """
    Return a list of IPs for current active network interfaces.
    """
    plat = platform.uname()[0]

    if plat == 'Windows':
        dirs = ['', r'c:\windows\system32', r'c:\winnt\system32']
        try:
            import ctypes
            buffer = ctypes.create_string_buffer(300)
            ctypes.windll.kernel32.GetSystemDirectoryA(buffer, 300)  # @UndefinedVariable
            dirs.insert(0, buffer.value.decode('mbcs'))
        except:
            pass
        for sysdir in dirs:
            try:
                pipe = os.popen(os.path.join(sysdir, 'ipconfig') + ' /all')
            except IOError:
                return []
            rawtxt = six.text_type(pipe.read())
            ips_unicode = re.findall(u'^.*?IP.*?(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}).*$', rawtxt, re.U | re.M)  # @UndefinedVariable
            ips = []
            for ip in ips_unicode:
                ips.append(str(ip))
            del ips_unicode
            return ips

    elif plat == 'Linux':
        if sys.executable == 'android_python' or ('ANDROID_ARGUMENT' in os.environ or 'ANDROID_ROOT' in os.environ):
            # fake nw interface on Android
            return [
                '127.0.0.1',
            ]

        try:
            pipe = os.popen('`which ip` -f inet a')
        except IOError:
            return []
        try:
            rawtxt = six.text_type(pipe.read())
            lines = rawtxt.splitlines()
        except:
            return []
        ips = set()
        for line in lines:
            check = line.strip('\n').strip().split(' ')
            if check[0] == 'inet':
                if check[2] == 'brd':
                    check.pop(2)
                    check.pop(2)
                ipaddress = check[1].split('/')[0]
                ips.add(str(ipaddress))
        return list(ips)

    elif plat == 'Darwin':
        try:
            # TODO: try to avoid socket connect to remote host
            return [
                _f for _f in [
                    l for l in (
                        [ip for ip in socket.gethostbyname_ex(socket.gethostname())[2] if not ip.startswith('127.')][:1],
                        # TODO: replace 8.8.8.8 with random seed node
                        [[(s.connect(('8.8.8.8', 53)), s.getsockname()[0], s.close()) for s in [socket.socket(socket.AF_INET, socket.SOCK_DGRAM)]][0][1]],
                    ) if l
                ][0] if _f
            ]
        except:
            eth0 = getIfconfig('eth0')
            en0 = getIfconfig('en0')
            return [_f for _f in [
                en0 or eth0,
            ] if _f]



#!/usr/bin/python
# contactsdb.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (contactsdb.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: contactsdb.

A low level methods to store list of contacts locally.:
    + suppliers
    + customers
    + correspondents
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from six.moves import range  # @UnresolvedImport

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import os

from twisted.internet.defer import DeferredList

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import nameurl
from bitdust.lib import strng
from bitdust.lib import jsn

from bitdust.system import bpio
from bitdust.system import local_fs

from bitdust.main import settings
from bitdust.main import listeners

from bitdust.services import driver

from bitdust.userid import id_url
from bitdust.userid import my_id
from bitdust.userid import global_id

from bitdust.contacts import identitycache

#-------------------------------------------------------------------------------

_CustomersList = []  # comes from settings.CustomerIDsFilename()
_SuppliersList = {}  # comes from settings.SuppliersDir()
_CorrespondentsList = []  # comes from settings.CorrespondentIDsFilename()
_CorrespondentsDict = {}

_SuppliersChangedCallback = None
_CustomersChangedCallback = None
_CorrespondentsChangedCallback = None
_ContactsChangedCallbacks = []

_CustomersMetaInfo = {}
_SuppliersMetaInfo = {}
_CorrespondentsMetaInfo = {}

#-------------------------------------------------------------------------------


def init():
    """
    We read from disk and if we have all the info we are set.

    If we don't have enough, then we have to ask BitDust to list
    contacts and use that list to get and then store all the identities
    for our contacts.
    """
    global _SuppliersChangedCallback
    global _CustomersChangedCallback
    global _CorrespondentsChangedCallback
    if _Debug:
        lg.out(_DebugLevel, 'contactsdb.init')
    d = cache_contacts()
    d.addCallback(lambda _: load_contacts())
    d.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='contactsdb.init')
    return d


def shutdown():
    global _SuppliersChangedCallback
    global _CustomersChangedCallback
    if _Debug:
        lg.out(_DebugLevel, 'contactsdb.shutdown')
    RemoveContactsChangedCallback(on_contacts_changed)
    if _SuppliersChangedCallback is not None:
        _SuppliersChangedCallback = None
    if _CustomersChangedCallback is not None:
        _CustomersChangedCallback = None


#------------------------------------------------------------------------------


def suppliers(customer_idurl=None):
    """
    Return list of suppliers ID's for given customer - me or another user.
    """
    global _SuppliersList
    if not customer_idurl:
        customer_idurl = my_id.getIDURL()
    customer_idurl = id_url.field(customer_idurl)
    if customer_idurl not in _SuppliersList:
        _SuppliersList[customer_idurl] = []
        lg.info('created new suppliers list in memory for customer %r' % customer_idurl)
    return _SuppliersList[customer_idurl]


def supplier(index, customer_idurl=None):
    """
    Return supplier ID on given position or empty string.
    """
    if not customer_idurl:
        customer_idurl = my_id.getIDURL()
    customer_idurl = id_url.field(customer_idurl)
    num = int(index)
    if num >= 0 and num < len(suppliers(customer_idurl=customer_idurl)):
        return suppliers(customer_idurl=customer_idurl)[num]
    return id_url.field(b'')


def all_suppliers(as_dict=False):
    global _SuppliersList
    if as_dict:
        return _SuppliersList
    result = []
    for suppliers_list in _SuppliersList.values():
        for supplier_idurl in suppliers_list:
            if id_url.is_cached(supplier_idurl) and supplier_idurl not in result:
                result.append(supplier_idurl)
    return result


def set_suppliers(idlist, customer_idurl=None):
    """
    Set suppliers ID's list.
    """
    global _SuppliersList
    if not customer_idurl:
        customer_idurl = my_id.getIDURL()
    customer_idurl = id_url.field(customer_idurl)
    if customer_idurl not in _SuppliersList:
        _SuppliersList[customer_idurl] = []
        lg.info('created new suppliers list in memory for customer %r' % customer_idurl)
    try:
        _SuppliersList[customer_idurl] = id_url.fields_list(idlist)
    except:
        pass
    if _Debug:
        lg.args(_DebugLevel, suppliers=_SuppliersList[customer_idurl], customer_idurl=customer_idurl)


def update_suppliers(idlist, customer_idurl=None):
    """
    High-level method to set suppliers ID's list.
    Executes required callbacks.
    """
    global _SuppliersChangedCallback
    global _ContactsChangedCallbacks
    oldsuppliers = list(suppliers(customer_idurl=customer_idurl))
    oldcontacts = list(contacts())
    idlist = list(map(lambda i: i if id_url.is_cached(i) else b'', idlist))
    set_suppliers(idlist, customer_idurl=customer_idurl)
    if _SuppliersChangedCallback is not None:
        _SuppliersChangedCallback(oldsuppliers, suppliers(customer_idurl=customer_idurl))
    if id_url.to_original_list(oldcontacts) != id_url.to_original_list(contacts()):
        for cb in _ContactsChangedCallbacks:
            cb(id_url.to_original_list(oldcontacts), id_url.to_original_list(contacts()))


def add_supplier(idurl, position=None, customer_idurl=None):
    """
    Add supplier in my list of suppliers or to the list stored for another customer.
    If parameter `position` is provided, supplier will be inserted instead of added.
    If position is greater than current list - empty strings will be filled in between.
    """
    global _SuppliersList
    if not customer_idurl:
        customer_idurl = my_id.getIDURL()
    customer_idurl = id_url.field(customer_idurl)
    if customer_idurl not in _SuppliersList:
        _SuppliersList[customer_idurl] = []
        lg.info('created new suppliers list in memory for customer %r' % customer_idurl)
    idurl = id_url.field(idurl)
    if _Debug:
        lg.args(_DebugLevel, idurl=idurl, position=position, customer_idurl=customer_idurl)
    if position is None or position == -1:
        lg.warn('position unknown, added supplier "%s" to the end of the list for customer %s' % (idurl, customer_idurl))
        _SuppliersList[customer_idurl].append(idurl)
        return len(_SuppliersList[customer_idurl]) - 1
    current_suppliers = _SuppliersList[customer_idurl]
    if position >= len(current_suppliers):
        empty_suppliers = [
            id_url.field(b''),
        ]*(1 + position - len(current_suppliers))
        current_suppliers.extend(empty_suppliers)
        if _Debug:
            lg.out(_DebugLevel, 'contactsdb.add_supplier   %d empty suppliers added for customer %r' % (len(empty_suppliers), customer_idurl))
    if current_suppliers[position] and current_suppliers[position] != idurl:
        lg.info('replacing known supplier "%s" by "%s" at position %d for customer %s' % (current_suppliers[position], idurl, position, customer_idurl))
    else:
        lg.info('added supplier "%s" at position %d for customer %s' % (idurl, position, customer_idurl))
    current_suppliers[position] = idurl
    update_suppliers(idlist=current_suppliers, customer_idurl=customer_idurl)
    return position


def erase_supplier(idurl=None, position=None, customer_idurl=None):
    global _SuppliersList
    if not customer_idurl:
        customer_idurl = my_id.getIDURL()
    customer_idurl = id_url.field(customer_idurl)
    if customer_idurl not in _SuppliersList:
        return False
    current_suppliers = _SuppliersList[customer_idurl]
    if _Debug:
        lg.args(_DebugLevel, idurl=idurl, position=position, customer_idurl=customer_idurl)
    if idurl:
        idurl = id_url.field(idurl)
        if idurl not in current_suppliers:
            return False
        current_suppliers[current_suppliers.index(idurl)] = id_url.field(b'')
    elif position is not None:
        if position >= len(current_suppliers):
            return False
        current_suppliers[position] = id_url.field(b'')
    else:
        return False
    update_suppliers(idlist=current_suppliers, customer_idurl=customer_idurl)
    return True


def clear_suppliers(customer_idurl=None):
    """
    Remove all suppliers for given customer, if customer_idurl is None will erase all my suppliers.
    """
    global _SuppliersList
    if not customer_idurl:
        customer_idurl = my_id.getIDURL()
    customer_idurl = id_url.field(customer_idurl)
    _SuppliersList.pop(customer_idurl)
    lg.info('erased suppliers list from memory for customer %r' % customer_idurl)


def clear_all_suppliers():
    """
    Remove all suppliers.
    """
    global _SuppliersList
    _SuppliersList.clear()


#------------------------------------------------------------------------------


def known_customers():
    """
    Return list of all known customers : if I need to connect to their suppliers.
    """
    global _SuppliersList
    return list(_SuppliersList.keys())


def customers():
    """
    Return list of my customers ID's if I am a supplier.
    """
    global _CustomersList
    return _CustomersList


def customer(index):
    """
    Return customer IDURL at given position or empty string if I have no customers.
    """
    num = int(index)
    if num >= 0 and num < len(customers()):
        return customers()[num]
    return id_url.field(b'')


def set_customers(idlist):
    """
    Set customers list.
    """
    global _CustomersList
    try:
        _CustomersList = id_url.fields_list(idlist)
    except:
        pass


def update_customers(idslist):
    """
    High-level method to set customers ID's list.
    Executes required callbacks.
    """
    global _CustomersChangedCallback
    global _ContactsChangedCallbacks
    old_customers = list(customers())
    old_contacts = list(contacts())
    idslist = list(filter(id_url.is_cached, idslist))
    set_customers(idslist)
    if _CustomersChangedCallback is not None:
        _CustomersChangedCallback(old_customers, customers())
    for cb in _ContactsChangedCallbacks:
        cb(id_url.to_original_list(old_contacts), id_url.to_original_list(contacts()))
    if _Debug:
        lg.args(_DebugLevel, new_customers=idslist, old_customers=old_customers)


def clear_customers():
    """
    Remove all customers.
    """
    global _CustomersList
    _CustomersList = []


#------------------------------------------------------------------------------


def contacts(include_all=False, include_enabled=True):
    """
    Return a union of suppliers and customers ID's.
    """
    result = set()
    if include_all or (include_enabled and driver.is_enabled('service_customer')) or driver.is_on('service_customer'):
        result.update(set(all_suppliers()))
    if include_all or (include_enabled and driver.is_enabled('service_supplier')) or driver.is_on('service_supplier'):
        result.update(set(customers() + known_customers()))
    if include_all or (include_enabled and driver.is_enabled('service_private_messages')) or driver.is_on('service_private_messages'):
        result.update(set(correspondents_ids()))
    if include_all or include_enabled:
        if driver.is_enabled('service_joint_postman') or driver.is_on('service_joint_postman'):
            from bitdust.stream import postman
            result.update(set(postman.list_customers()))
            result.update(set(postman.list_consumers_producers(include_consumers=True, include_producers=True)))
    return list(result)


def contacts_list():
    """
    Return a list of suppliers and customers ID's.
    """
    return list(suppliers() + customers())


def contacts_remote(include_all=False, include_enabled=True):
    """
    Return ID's list of all known peers.
    """
    l = id_url.to_bin_list(contacts(include_all=include_all, include_enabled=include_enabled))
    return [i for i in id_url.fields_list(l) if not id_url.is_the_same(i, my_id.getIDURL())]


#------------------------------------------------------------------------------


def correspondents():
    """
    Return list of tuples of correspondents: (IDURL, nickname).
    """
    global _CorrespondentsList
    return _CorrespondentsList


def correspondents_ids():
    """
    Return list of correspondents IDURLs.
    """
    global _CorrespondentsList
    return [tupl[0] for tupl in _CorrespondentsList]


def correspondents_dict():
    """
    Return dictionary of correspondents : IDURL->nickname.
    """
    global _CorrespondentsList
    return dict(_CorrespondentsList)


def set_correspondents(idlist):
    """
    Set correspondents from list of tuples without notification.
    """
    global _CorrespondentsList
    _CorrespondentsList = [(id_url.field(idurl_name[0]).to_bin(), idurl_name[1]) for idurl_name in idlist]


def add_correspondent(idurl, nickname=''):
    """
    Add correspondent, execute notification callback and return its position in
    the list.
    """
    global _CorrespondentsList
    global _CorrespondentsChangedCallback
    curlist = list(_CorrespondentsList)
    idurl = id_url.field(idurl)
    _CorrespondentsList.append((
        idurl.to_bin(),
        nickname,
    ))
    if _CorrespondentsChangedCallback is not None:
        _CorrespondentsChangedCallback(curlist, _CorrespondentsList)
    listeners.push_snapshot('correspondent', snap_id=idurl.to_bin(), data=dict(
        idurl=idurl.to_bin(),
        nickname=nickname,
    ))
    return len(curlist)


def remove_correspondent(idurl):
    """
    Remove correspondent with given IDURL, execute notification callback and
    return True if success.
    """
    global _CorrespondentsList
    global _CorrespondentsChangedCallback
    curlist = list(_CorrespondentsList)
    idurl = id_url.field(idurl)
    for tupl in _CorrespondentsList:
        if idurl.to_bin() == id_url.field(tupl[0]).to_bin():
            _CorrespondentsList.remove(tupl)
            if _CorrespondentsChangedCallback is not None:
                _CorrespondentsChangedCallback(curlist, _CorrespondentsList)
            listeners.push_snapshot('correspondent', snap_id=idurl.to_bin(), deleted=True, data=dict(
                idurl=idurl.to_bin(),
                nickname=tupl[1],
            ))
            return True
    return False


def populate_correspondents():
    for corr in correspondents():
        listeners.push_snapshot('correspondent', snap_id=corr[0], data=dict(
            idurl=corr[0],
            nickname=corr[1],
        ))


#-------------------------------------------------------------------------------


def is_customer(idurl):
    """
    Return True if given ID is found in customers list.
    """
    if id_url.is_empty(idurl):
        return False
    return id_url.field(idurl).to_bin() in id_url.to_bin_list(customers())


def is_supplier(idurl, customer_idurl=None):
    """
    Return True if given ID is found in suppliers list.
    """
    if id_url.is_empty(idurl):
        return False
    return id_url.field(idurl).to_bin() in id_url.to_bin_list(suppliers(customer_idurl=customer_idurl))


def is_correspondent(idurl):
    """
    Return True if given ID is found in correspondents list.
    """
    if id_url.is_empty(idurl):
        return False
    return id_url.field(idurl).to_bin() in id_url.to_bin_list(correspondents_ids())


#------------------------------------------------------------------------------


def num_customers():
    """
    Return current number of customers.
    """
    return len(customers())


def num_suppliers(customer_idurl=None):
    """
    Return current number of suppliers.
    """
    return len(suppliers(customer_idurl=customer_idurl))


def total_suppliers():
    global _SuppliersList
    result = set()
    for suppliers_list in _SuppliersList.values():
        result.update(set(id_url.to_bin_list(suppliers_list)))
    return len(result)


def num_correspondents():
    """
    Return current number of correspondents.
    """
    return len(correspondents())


#------------------------------------------------------------------------------


def supplier_position(idurl, customer_idurl=None):
    """
    Return position of supplier with given ID or -1.
    """
    if not idurl:
        return -1
    if not id_url.is_cached(idurl):
        return -1
    idurl = id_url.field(idurl)
    try:
        index = id_url.to_bin_list(suppliers(customer_idurl=customer_idurl)).index(idurl.to_bin())
    except:
        index = -1
    return index


def customer_position(idurl):
    """
    Return position of supplier with given ID or -1.
    """
    if not idurl:
        return -1
    if not id_url.is_cached(idurl):
        return -1
    idurl = id_url.field(idurl)
    try:
        index = id_url.to_bin_list(customers()).index(idurl.to_bin())
    except:
        index = -1
    return index


def contact_position(idurl):
    """
    Return position for given contact ID in the total list combined from
    suppliers, customers.

    Suppliers should be numbered 0 to 63 with customers after that not
    sure we can count on numbers staying.
    """
    if not idurl:
        return -1
    if not id_url.is_cached(idurl):
        return -1
    idurl = id_url.field(idurl)
    try:
        index = id_url.to_bin_list(contacts_list()).index(idurl.to_bin())
    except:
        index = -1
    return index


#-------------------------------------------------------------------------------


def save_suppliers(path=None, customer_idurl=None):
    """
    Write current suppliers list on the disk, ``path`` is a file path to save.
    """
    if not customer_idurl:
        customer_idurl = my_id.getIDURL()
    customer_idurl = id_url.field(customer_idurl)
    customer_id = global_id.UrlToGlobalID(customer_idurl)
    if path is None:
        path = os.path.join(settings.SuppliersDir(), customer_id, 'supplierids')
    lst = suppliers(customer_idurl=customer_idurl)
    lst = list(map(strng.to_text, lst))
    if not os.path.exists(os.path.dirname(path)):
        bpio._dirs_make(os.path.dirname(path))
    bpio._write_list(path, lst)
    if _Debug:
        lg.out(_DebugLevel, 'contactsdb.save_suppliers for customer [%s]:\n%r' % (customer_id, lst))
    return True


def load_suppliers(path=None, customer_idurl=None, all_customers=False):
    """
    Load suppliers list from disk.
    """
    if all_customers:
        list_local_customers = list(os.listdir(settings.SuppliersDir()))
        if _Debug:
            lg.out(_DebugLevel, 'contactsdb.load_suppliers %d known customers' % len(list_local_customers))
        for customer_id in list_local_customers:
            if not global_id.IsValidGlobalUser(customer_id):
                lg.warn('invalid customer record %s found in %s' % (customer_id, settings.SuppliersDir()))
                continue
            path = os.path.join(settings.SuppliersDir(), customer_id, 'supplierids')
            lst = bpio._read_list(path)
            if lst is None:
                lg.warn('did not found suppliers ids at %s' % path)
                continue
            one_customer_idurl = global_id.GlobalUserToIDURL(customer_id)
            if not id_url.is_cached(one_customer_idurl):
                lg.warn('customer identity %r not cached yet' % one_customer_idurl)
                continue
            if not one_customer_idurl.is_latest():
                latest_customer_path = os.path.join(settings.SuppliersDir(), one_customer_idurl.to_id())
                old_customer_path = os.path.join(settings.SuppliersDir(), customer_id)
                if not os.path.exists(latest_customer_path):
                    os.rename(old_customer_path, latest_customer_path)
                    lg.info('detected and processed idurl rotate when loading suppliers for customer : %r -> %r' % (customer_id, one_customer_idurl.to_id()))
                else:
                    bpio._dir_remove(old_customer_path)
                    lg.warn('found old customer dir %r and removed' % old_customer_path)
                    continue
            lst = list(map(lambda i: i if id_url.is_cached(i) else b'', lst))
            set_suppliers(lst, customer_idurl=one_customer_idurl)
            if _Debug:
                lg.out(_DebugLevel, '    loaded %d known suppliers for customer %r' % (len(lst), one_customer_idurl))
        return True
    if not customer_idurl:
        customer_idurl = my_id.getIDURL()
    customer_idurl = id_url.field(customer_idurl)
    if path is None:
        path = os.path.join(settings.SuppliersDir(), global_id.UrlToGlobalID(customer_idurl), 'supplierids')
    lst = bpio._read_list(path)
    if lst is None:
        lst = list()
    lst = list(map(lambda i: i if id_url.is_cached(i) else b'', lst))
    set_suppliers(lst, customer_idurl=customer_idurl)
    if _Debug:
        lg.out(_DebugLevel, 'contactsdb.load_suppliers %d items from %s' % (len(lst), path))
    return True


def cache_suppliers(path=None):
    """
    Make sure identities of all suppliers we know are cached.
    """
    dl = []
    list_local_customers = list(os.listdir(settings.SuppliersDir()))
    for customer_id in list_local_customers:
        if not global_id.IsValidGlobalUser(customer_id):
            lg.warn('invalid customer record %s found in %s' % (customer_id, settings.SuppliersDir()))
            continue
        try:
            one_customer_idurl = global_id.GlobalUserToIDURL(customer_id)
        except Exception as exc:
            lg.err('idurl caching failed: %r' % exc)
            continue
        if not id_url.is_cached(one_customer_idurl):
            dl.append(identitycache.immediatelyCaching(one_customer_idurl))
        path = os.path.join(settings.SuppliersDir(), customer_id, 'supplierids')
        lst = bpio._read_list(path)
        if lst is None:
            lg.warn('did not found suppliers ids at %s' % path)
            continue
        for one_supplier_idurl in lst:
            if one_supplier_idurl:
                if not id_url.is_cached(one_supplier_idurl):
                    dl.append(identitycache.immediatelyCaching(one_supplier_idurl))
    if _Debug:
        lg.out(_DebugLevel, 'contactsdb.cache_suppliers prepared %d idurls to be cached' % len(dl))
    return DeferredList(dl, consumeErrors=True)


#------------------------------------------------------------------------------


def save_customers(path=None, save_meta_info=False):
    """
    Write current customers list on the disk, ``path`` is a file path to save.
    """
    global _CustomersMetaInfo
    if path is None:
        path = settings.CustomerIDsFilename()
    lst = customers()
    lst = list(map(strng.to_text, lst))
    bpio._write_list(path, lst)
    if save_meta_info:
        json_info = id_url.to_bin_dict(_CustomersMetaInfo)
        local_fs.WriteTextFile(settings.CustomersMetaInfoFilename(), jsn.dumps(
            json_info,
            indent=2,
            sort_keys=True,
            keys_to_text=True,
        ))
    if _Debug:
        lg.out(_DebugLevel, 'contactsdb.save_customers save_meta_info=%r : %r' % (save_meta_info, lst))


def load_customers(path=None):
    """
    Load customers list from disk.
    """
    global _CustomersMetaInfo
    if path is None:
        path = settings.CustomerIDsFilename()
    lst = bpio._read_list(path)
    if lst is None:
        lst = list()
    lst = list(filter(id_url.is_cached, lst))
    set_customers(lst)
    _CustomersMetaInfo = jsn.loads(
        local_fs.ReadTextFile(settings.CustomersMetaInfoFilename()) or '{}',
        keys_to_bin=True,
    )
    _CustomersMetaInfo = id_url.to_bin_dict(_CustomersMetaInfo)
    _CustomersMetaInfo = jsn.dict_values_to_text(_CustomersMetaInfo)
    if _Debug:
        lg.out(_DebugLevel, 'contactsdb.load_customers %d items' % len(lst))


def cache_customers(path=None):
    """
    Make sure identities of all customers we know are cached.
    """
    dl = []
    if path is None:
        path = settings.CustomerIDsFilename()
    lst = bpio._read_list(path) or []
    for one_customer_idurl in lst:
        if one_customer_idurl:
            if not id_url.is_cached(one_customer_idurl):
                dl.append(identitycache.immediatelyCaching(one_customer_idurl))
    if _Debug:
        lg.out(_DebugLevel, 'contactsdb.cache_customers prepared %d idurls to be cached' % len(dl))
    return DeferredList(dl, consumeErrors=True)


#------------------------------------------------------------------------------


def save_correspondents(path=None):
    """
    Write current correspondents list on the disk, ``path`` is a file path to
    save.
    """
    if path is None:
        path = settings.CorrespondentIDsFilename()
    lst = ['%s %s' % (
        strng.to_text(t[0]),
        strng.to_text(t[1]),
    ) for t in correspondents()]
    bpio._write_list(path, lst)


def load_correspondents(path=None):
    """
    Load correspondents list from disk.
    """
    if path is None:
        path = settings.CorrespondentIDsFilename()
    lst = bpio._read_list(path)
    if lst is None:
        lst = list()
    for i in range(len(lst)):
        lst[i] = tuple(lst[i].strip().split(' ', 1))
        if len(lst[i]) < 2:
            lst[i] = (lst[i][0], '')
        if not lst[i][1].strip():
            lst[i] = (id_url.field(lst[i][0]), nameurl.GetName(lst[i][0]))
    lst = list(filter(lambda i: id_url.is_cached(i[0]), lst))
    set_correspondents(lst)
    if _Debug:
        lg.out(_DebugLevel, 'contactsdb.load_correspondents %d items' % len(lst))


def cache_correspondents(path=None):
    """
    Make sure identities of all correspondents we know are cached.
    """
    dl = []
    if path is None:
        path = settings.CorrespondentIDsFilename()
    lst = bpio._read_list(path) or []
    for i in range(len(lst)):
        try:
            one_correspondent_idurl = lst[i].strip().split(' ', 1)[0]
        except:
            lg.exc()
            continue
        if one_correspondent_idurl:
            if not id_url.is_cached(one_correspondent_idurl):
                dl.append(identitycache.immediatelyCaching(one_correspondent_idurl))
    if _Debug:
        lg.out(_DebugLevel, 'contactsdb.cache_correspondents prepared %d idurls to be cached' % len(dl))
    return DeferredList(dl, consumeErrors=True)


#------------------------------------------------------------------------------


def cache_contacts(
    suppliers_path=None,
    customers_path=None,
    correspondents_path=None,
):
    """
    Make sure identities of all my known contacts are cached.
    """
    dl = []
    dl.append(cache_suppliers(suppliers_path))
    dl.append(cache_customers(customers_path))
    dl.append(cache_correspondents(correspondents_path))
    return DeferredList(dl, consumeErrors=True)


def load_contacts():
    """
    Load all my contacts from disk.
    """
    load_suppliers(all_customers=True)
    if _SuppliersChangedCallback is not None:
        _SuppliersChangedCallback([], suppliers())
    load_customers()
    if _CustomersChangedCallback is not None:
        _CustomersChangedCallback([], customers())
    load_correspondents()
    if _CorrespondentsChangedCallback is not None:
        _CorrespondentsChangedCallback([], correspondents())
    AddContactsChangedCallback(on_contacts_changed)
    if listeners.is_populate_required('correspondent'):
        # listeners.populate_later().remove('correspondent')
        populate_correspondents()


#------------------------------------------------------------------------------


def get_contact_identity(idurl):
    """
    The Main Method Here - return identity object for given ID or None if not found.
    Only valid contacts for packets will be signed by local identity, suppliers, customers.
    """
    if idurl is None:
        return None
    idurl = id_url.field(idurl)
    if idurl.to_bin() == my_id.getIDURL().to_bin():
        return my_id.getLocalIdentity()


#     if is_supplier(idurl):
#         return identitycache.FromCache(idurl)
#     if is_customer(idurl):
#         return identitycache.FromCache(idurl)
#     if is_correspondent(idurl):
#         return identitycache.FromCache(idurl)
    if identitycache.HasKey(idurl):
        # lg.warn("who is %s ?" % nameurl.GetName(idurl))
        return identitycache.FromCache(idurl)
    lg.warn('%s is NOT FOUND IN CACHE' % idurl)
    # TODO:
    # this is not correct:
    # need to check if other contacts is fine - if internet is turned off we can get lots of fails ...
    return None


def get_customer_identity(idurl):
    """
    If ``idurl`` is in customers list, return its identity object.
    """
    if is_customer(idurl):
        idurl = id_url.field(idurl)
        return identitycache.FromCache(idurl)
    return None


def get_supplier_identity(idurl):
    """
    Return peer's identity if he is in suppliers list.
    """
    if is_supplier(idurl):
        idurl = id_url.field(idurl)
        return identitycache.FromCache(idurl)
    return None


def get_correspondent_identity(idurl):
    """
    Return peer's identity if he is in the correspondents list.
    """
    if is_correspondent(idurl):
        idurl = id_url.field(idurl)
        return identitycache.FromCache(idurl)
    return None


def get_correspondent_nickname(correspondent_idurl):
    for idurl, nickname in correspondents():
        if id_url.field(idurl).to_bin() == id_url.field(correspondent_idurl).to_bin():
            return nickname
    return None


def find_correspondent_by_nickname(nickname):
    for idurl, corr_nickname in correspondents_dict().items():
        if nickname == corr_nickname:
            return idurl
    return None


#------------------------------------------------------------------------------


def on_contacts_changed(old_contacts_list, new_contacts_list):
    from bitdust.main import events
    events.send('contacts-changed', data=dict(
        old_contacts=old_contacts_list,
        new_contacts=new_contacts_list,
    ))


#------------------------------------------------------------------------------


def read_customers_meta_info_all():
    global _CustomersMetaInfo
    return _CustomersMetaInfo


def write_customers_meta_info_all(new_customers_info):
    global _CustomersMetaInfo
    _CustomersMetaInfo = new_customers_info
    json_info = {k: jsn.dict_keys_to_text(v) for k, v in id_url.to_bin_dict(_CustomersMetaInfo).items()}
    try:
        raw_data = jsn.dumps(
            json_info,
            indent=2,
            sort_keys=True,
            keys_to_text=True,
            values_to_text=True,
        )
    except:
        lg.exc()
        return None
    local_fs.WriteTextFile(settings.CustomersMetaInfoFilename(), raw_data)
    return _CustomersMetaInfo


def add_customer_meta_info(customer_idurl, info):
    global _CustomersMetaInfo
    customer_idurl = id_url.field(customer_idurl)
    if not customer_idurl.is_latest():
        if customer_idurl.original() in _CustomersMetaInfo:
            if customer_idurl.to_bin() not in _CustomersMetaInfo:
                _CustomersMetaInfo[customer_idurl.to_bin()] = _CustomersMetaInfo.pop(customer_idurl.original())
                lg.info('detected and processed idurl rotate for customer meta info : %r -> %r' % (customer_idurl.original(), customer_idurl.to_bin()))
    customer_idurl = id_url.to_bin(customer_idurl)
    if 'family_snapshot' in info:
        info['family_snapshot'] = id_url.to_bin_list(info['family_snapshot'])
    if 'ecc_map' in info:
        info['ecc_map'] = strng.to_text(info['ecc_map'])
    if customer_idurl not in _CustomersMetaInfo:
        if _Debug:
            lg.out(_DebugLevel, 'contactsdb.add_customer_meta_info   store new meta info for customer %r: %r' % (customer_idurl, info))
        _CustomersMetaInfo[customer_idurl] = {}
    else:
        if _Debug:
            lg.out(_DebugLevel, 'contactsdb.add_customer_meta_info   update existing meta info for customer %r: %r' % (customer_idurl, info))
        _CustomersMetaInfo[customer_idurl].update(info)
    json_info = {k: jsn.dict_keys_to_text(v) for k, v in id_url.to_bin_dict(_CustomersMetaInfo).items()}
    try:
        raw_data = jsn.dumps(
            json_info,
            indent=2,
            sort_keys=True,
            keys_to_text=True,
            values_to_text=True,
        )
    except:
        lg.exc()
        return None
    local_fs.WriteTextFile(settings.CustomersMetaInfoFilename(), raw_data)
    return _CustomersMetaInfo


def remove_customer_meta_info(customer_idurl):
    global _CustomersMetaInfo
    customer_idurl = id_url.field(customer_idurl)
    if not customer_idurl.is_latest():
        if customer_idurl.original() in _CustomersMetaInfo:
            if customer_idurl.to_bin() not in _CustomersMetaInfo:
                _CustomersMetaInfo[customer_idurl.to_bin()] = _CustomersMetaInfo.pop(customer_idurl.original())
                lg.info('detected and processed idurl rotate for customer meta info : %r -> %r' % (customer_idurl.original(), customer_idurl.to_bin()))
    customer_idurl = id_url.to_bin(customer_idurl)
    if customer_idurl not in _CustomersMetaInfo:
        lg.warn('meta info for customer %r not exist' % customer_idurl)
        return False
    if _Debug:
        lg.out(_DebugLevel, 'contactsdb.remove_customer_meta_info   erase existing meta info for customer %r' % customer_idurl)
    _CustomersMetaInfo.pop(customer_idurl)
    json_info = {k: jsn.dict_keys_to_text(v) for k, v in id_url.to_bin_dict(_CustomersMetaInfo).items()}
    local_fs.WriteTextFile(settings.CustomersMetaInfoFilename(), jsn.dumps(
        json_info,
        indent=2,
        sort_keys=True,
        keys_to_text=True,
        values_to_text=True,
    ))
    return True


def get_customer_meta_info(customer_idurl):
    global _CustomersMetaInfo
    customer_idurl = id_url.field(customer_idurl)
    if not customer_idurl.is_latest():
        if customer_idurl.original() in _CustomersMetaInfo:
            if customer_idurl.to_bin() not in _CustomersMetaInfo:
                _CustomersMetaInfo[customer_idurl.to_bin()] = _CustomersMetaInfo.pop(customer_idurl.original())
                lg.info('detected and processed idurl rotate for customer meta info : %r -> %r' % (customer_idurl.original(), customer_idurl.to_bin()))
    customer_idurl = id_url.to_bin(customer_idurl)
    return jsn.dict_keys_to_text(jsn.dict_values_to_text(_CustomersMetaInfo.get(customer_idurl, {})))


#------------------------------------------------------------------------------


def add_supplier_meta_info(supplier_idurl, info, customer_idurl=None):
    global _SuppliersMetaInfo
    if not customer_idurl:
        customer_idurl = my_id.getIDURL()
    customer_idurl = id_url.field(customer_idurl)
    supplier_idurl = id_url.field(supplier_idurl)
    if customer_idurl not in _SuppliersMetaInfo:
        _SuppliersMetaInfo[customer_idurl] = {}
    if supplier_idurl not in _SuppliersMetaInfo[customer_idurl]:
        _SuppliersMetaInfo[customer_idurl][supplier_idurl] = {}
    _SuppliersMetaInfo[customer_idurl][supplier_idurl].update(info)
    if _Debug:
        lg.out(_DebugLevel, 'contactsdb.add_supplier_meta_info   for supplier %s of customer %s: %s' % (supplier_idurl, customer_idurl, _SuppliersMetaInfo[customer_idurl][supplier_idurl]))


def remove_supplier_meta_info(supplier_idurl, customer_idurl=None):
    global _SuppliersMetaInfo
    if not customer_idurl:
        customer_idurl = my_id.getIDURL()
    customer_idurl = id_url.field(customer_idurl)
    supplier_idurl = id_url.field(supplier_idurl)
    if customer_idurl not in _SuppliersMetaInfo:
        return False
    if supplier_idurl not in _SuppliersMetaInfo[customer_idurl]:
        return False
    _SuppliersMetaInfo[customer_idurl].pop(supplier_idurl)
    if len(_SuppliersMetaInfo[customer_idurl]) == 0:
        _SuppliersMetaInfo.pop(customer_idurl)
    return True


def get_supplier_meta_info(supplier_idurl, customer_idurl=None):
    global _SuppliersMetaInfo
    if not customer_idurl:
        customer_idurl = my_id.getIDURL()
    if not id_url.is_cached(customer_idurl) or not id_url.is_cached(supplier_idurl):
        return {}
    customer_idurl = id_url.field(customer_idurl)
    supplier_idurl = id_url.field(supplier_idurl)
    return jsn.dict_keys_to_text(jsn.dict_values_to_text(_SuppliersMetaInfo.get(customer_idurl, {}).get(supplier_idurl, {})))


#------------------------------------------------------------------------------


def SetSuppliersChangedCallback(cb):
    """
    Set callback to fire when suppliers is changed.
    """
    global _SuppliersChangedCallback
    _SuppliersChangedCallback = cb


def SetCustomersChangedCallback(cb):
    """
    Set callback to fire when customers is changed.
    """
    global _CustomersChangedCallback
    _CustomersChangedCallback = cb


def SetCorrespondentsChangedCallback(cb):
    """
    Set callback to fire when correspondents is changed.
    """
    global _CorrespondentsChangedCallback
    _CorrespondentsChangedCallback = cb


def AddContactsChangedCallback(cb):
    """
    Set callback to fire when any contact were changed.
    """
    global _ContactsChangedCallbacks
    _ContactsChangedCallbacks.append(cb)


def RemoveContactsChangedCallback(cb):
    """
    Set callback to fire when any contact were changed.
    """
    global _ContactsChangedCallbacks
    if cb in _ContactsChangedCallbacks:
        _ContactsChangedCallbacks.remove(cb)



#!/usr/bin/env python3
"""

Demo script for the api_getaddresssince api command.

takes
- last known block
- min confirmations
- address

Sends back at most 720 blocks worth of data with

- last considered block
- min confirmations
- list of tx matching the address

Usage:

python3 demo_getaddresssince.py block_height min_conf address

eg:
`python3 demo_getaddresssince.py 864115 20 edf2d63cdf0b6275ead22c9e6d66aa8ea31dc0ccb367fad2e7c08a25`


No proper error checking.
"""

import connections
import json
import socks
import sys

__version__ = '0.0.2'


def get_address_since(since, min_conf, address):
    s = socks.socksocket()
    s.settimeout(10)
    s.connect(('127.0.0.1', 5658))
    # Command first
    connections.send(s, 'api_getaddresssince')
    # Then last block (will not be included in results
    connections.send(s, int(since))
    # min confirmations
    connections.send(s, int(min_conf))
    # and finally the address
    connections.send(s, str(address))

    res = connections.receive(s)
    return res


if __name__ == '__main__':
    _, since, min_conf, address = sys.argv
    print('api_getaddresssince since {} minconf={} for address {}'.format(since, min_conf, address))
    res_as_native_dict = get_address_since(since, min_conf, address)
    print(json.dumps(res_as_native_dict))



#!/usr/bin/python
# service_ip_port_responder.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_ip_port_responder.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_ip_port_responder
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return IPPortResponderService()


class IPPortResponderService(LocalService):

    service_name = 'service_ip_port_responder'
    config_path = 'services/ip-port-responder/enabled'
    start_suspended = True

    def dependent_on(self):
        return [
            'service_udp_datagrams',
            'service_entangled_dht',
        ]

    def start(self):
        from bitdust.stun import stun_server
        from bitdust.main import settings
        from bitdust.lib import udp
        from bitdust.logs import lg
        if not udp.proto(settings.getUDPPort()):
            lg.warn('udp port %s is not opened yet' % settings.getUDPPort())
            return False
        udp_port = int(settings.getUDPPort())
        stun_server.A('start', udp_port)
        return True

    def stop(self):
        from bitdust.stun import stun_server
        stun_server.A('stop')
        stun_server.Destroy()
        return True

    def on_suspend(self, *args, **kwargs):
        from bitdust.stun import stun_server
        stun_server.A('stop')
        return True

    def on_resume(self, *args, **kwargs):
        from bitdust.stun import stun_server
        from bitdust.main import settings
        udp_port = int(settings.getUDPPort())
        stun_server.A('start', udp_port)
        return True



#!/usr/bin/python
# lg.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (lg.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: lg
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

import six
import os
import sys
import time
import datetime
import threading
import traceback
import logging
import platform
from io import open

#------------------------------------------------------------------------------

_GlobalDebugLevel = 0
_LogLinesCounter = 0
_LogsEnabled = True
_UseColors = None
_NoOutput = False
_IsAndroid = None
_InterceptedLogFile = None
_RedirectStdOut = False
_RedirectStdErr = False
_OriginalStdOut = None
_OriginalStdErr = None
_StdOutPrev = None
_StdErrPrev = None
_LogFile = None
_LogFileName = None
_AllLogFiles = {}
_StoreExceptionsEnabled = True
_WebStreamFunc = None
_ShowTime = True
_LifeBeginsTime = 0
_TimePush = 0.0
_TimeTotalDict = {}
_TimeDeltaDict = {}
_TimeCountsDict = {}

#------------------------------------------------------------------------------


def fqn(o):
    return o.__module__ + '.' + o.__name__


#------------------------------------------------------------------------------


def out(_DebugLevel, msg, nl='\n', log_name='stdout', showtime=False):
    """
    Prints a text line to the log file or console.

    :param level: lower values are counting as more important messages.
                        Usually I am using only even values from 0 to 18.
    :param msg: message string to be printed
    :param nl: this string is added at the end,
               set to empty string to avoid new line.
    """
    global _IsAndroid
    global _InterceptedLogFile
    global _WebStreamFunc
    global _LogFile
    global _LogFileName
    global _RedirectStdOut
    global _RedirectStdErr
    global _ShowTime
    global _LifeBeginsTime
    global _NoOutput
    global _LogLinesCounter
    global _LogsEnabled
    global _UseColors
    global _GlobalDebugLevel
    global _AllLogFiles
    s = msg
    s_ = s
    level = _DebugLevel
    if not level:
        level = 0
    if level < 0:
        level = 0
    if level % 2:
        level -= 1
    if level:
        lstr = s.lstrip()
        if lstr.startswith('INFO') or lstr.startswith('DEBUG') or lstr.startswith('DEBUG') or lstr.startswith('WARNING') or lstr.startswith('ERROR'):
            s = '  ' + lstr
        else:
            s = ' '*level + s
    if _IsAndroid is None:
        _IsAndroid = (sys.executable == 'android_python' or ('ANDROID_ARGUMENT' in os.environ or 'ANDROID_ROOT' in os.environ))
    if (_ShowTime and level > 0) or showtime:
        tm_string = time.strftime('%H:%M:%S')
        if _LifeBeginsTime != 0:
            dt = time.time() - _LifeBeginsTime
            mn = dt // 60
            sc = dt - mn*60
            if _GlobalDebugLevel >= 6:
                tm_string += '/%02d:%06.3f' % (mn, sc)
            else:
                tm_string += '/%02d:%02d' % (mn, sc)
        if _UseColors is None:
            _UseColors = platform.uname()[0] != 'Windows' and os.environ.get('BITDUST_LOG_USE_COLORS', '1') != '0'
        if _UseColors:
            tm_string = '\033[0;49;97m%s\033[0m' % tm_string
        if level == 0:
            tm_string += '  '
        s = tm_string + s
    if is_debug(30):
        currentThreadName = threading.currentThread().getName()
        s = s + ' {%s}' % currentThreadName.lower()
    if _InterceptedLogFile:
        if is_debug(level):
            _InterceptedLogFile.write(log_name + ': ' + s + nl)
            _InterceptedLogFile.flush()
        return
    if not _LogsEnabled:
        return
    if is_debug(level):
        if log_name == 'stdout':
            if _LogFile is not None:
                o = s + nl
                if sys.version_info[0] == 3:
                    if not isinstance(o, str):
                        o = o.decode('utf-8')
                else:
                    if not isinstance(o, unicode):  # @UndefinedVariable
                        o = o.decode('utf-8')
                try:
                    _LogFile.write(o)
                    _LogFile.flush()
                except:
                    pass
        else:
            if _LogFileName:
                if log_name not in _AllLogFiles:
                    filename = os.path.join(os.path.dirname(_LogFileName), log_name + '.log')
                    if not os.path.isdir(os.path.dirname(os.path.abspath(filename))):
                        os.makedirs(os.path.dirname(os.path.abspath(filename)))
                    _AllLogFiles[log_name] = open(os.path.abspath(filename), 'w')
                o = s + nl
                if sys.version_info[0] == 3:
                    if not isinstance(o, str):
                        o = o.decode('utf-8')
                else:
                    if not isinstance(o, unicode):  # @UndefinedVariable
                        o = o.decode('utf-8')
                try:
                    _AllLogFiles[log_name].write(o)
                    _AllLogFiles[log_name].flush()
                except:
                    pass
        if not _RedirectStdOut and not _RedirectStdErr and not _NoOutput:
            if log_name == 'stdout':
                s = s + nl
                try:
                    sys.stdout.write(s)
                except:
                    try:
                        sys.stdout.write(format_exception() + '\n\n' + s)
                    except:
                        # very bad stuff... we can't write anything to stdout?
                        pass
    if _WebStreamFunc is not None:
        _WebStreamFunc(level, s_ + nl)
    _LogLinesCounter += 1
    # if _LogLinesCounter % 10000 == 0:
    #     out(10, '[%s]' % time.asctime())
    return None


def dbg(_DebugLevel, message, *args, **kwargs):
    level = _DebugLevel
    cod = sys._getframe().f_back.f_code
    modul = os.path.basename(cod.co_filename).replace('.py', '')
    caller = cod.co_name
    funcname = '%s.%s' % (modul, caller)
    o = '%s %s' % (funcname, message)
    out(level, o, showtime=True)
    return o


def args(_DebugLevel, *args, **kwargs):
    level = _DebugLevel
    cod = sys._getframe().f_back.f_code
    modul = os.path.basename(cod.co_filename).replace('.py', '')
    caller = cod.co_name
    message = kwargs.pop('message', None)
    funcargs = []
    for k, v in enumerate(args):
        funcargs.append('%r' % v)
    for k in kwargs:
        funcargs.append('%s=%r' % (k, kwargs.get(k)))
    funcname = '%s.%s' % (modul, caller)
    o = '%s(%s)' % (
        funcname,
        ', '.join(funcargs),
    )
    if message:
        o += ' ' + message
    out(level, o, showtime=True)
    return o


def info(message, level=2):
    global _UseColors
    if _UseColors is None:
        _UseColors = platform.uname()[0] != 'Windows' and os.environ.get('BITDUST_LOG_USE_COLORS', '1') != '0'
    cod = sys._getframe().f_back.f_code
    modul = os.path.basename(cod.co_filename).replace('.py', '')
    caller = cod.co_name
    output_string = 'INFO %s in %s.%s()' % (
        message,
        modul,
        caller,
    )
    if _UseColors:
        output_string = '\033[0;49;92mINFO %s \033[0m\033[0;49;37min %s.%s()\033[0m' % (
            message,
            modul,
            caller,
        )
    out(level, output_string, showtime=True)
    # out(level, output_string, log_name='info', showtime=True)
    return message


def warn(message, level=2):
    global _UseColors
    if _UseColors is None:
        _UseColors = platform.uname()[0] != 'Windows' and os.environ.get('BITDUST_LOG_USE_COLORS', '1') != '0'
    cod = sys._getframe().f_back.f_code
    modul = os.path.basename(cod.co_filename).replace('.py', '')
    caller = cod.co_name
    output_string = 'WARNING %s in %s.%s()' % (
        message,
        modul,
        caller,
    )
    if _UseColors:
        output_string = '\033[0;35mWARNING %s \033[0m\033[0;49;37min %s.%s()\033[0m' % (
            message,
            modul,
            caller,
        )
    out(level, output_string, showtime=True)
    # out(level, output_string, log_name='warn', showtime=True)
    return message


def err(message, level=0):
    global _UseColors
    if _UseColors is None:
        _UseColors = platform.uname()[0] != 'Windows' and os.environ.get('BITDUST_LOG_USE_COLORS', '1') != '0'
    cod = sys._getframe().f_back.f_code
    modul = os.path.basename(cod.co_filename).replace('.py', '')
    caller = cod.co_name
    funcname = '%s.%s' % (modul, caller)
    if isinstance(message, Exception):
        message = str(message)
    if not isinstance(message, six.text_type):  # @UndefinedVariable
        message = str(message)
    if not message.count(funcname):
        message = ' %s in %s()' % (message, funcname)
    if not message.count('ERROR'):
        message = 'ERROR ' + message
    message = '%s%s   ' % ((' '*(level + 11)), message)
    if _UseColors:
        message = '\033[1;37;41m%s\033[0m' % message
    out(level, message, showtime=True)
    # out(level, message, log_name='err', showtime=True)
    return message


def exc(msg='', level=0, maxTBlevel=100, exc_info=None, exc_value=None, **kwargs):
    global _UseColors
    if _UseColors is None:
        _UseColors = platform.uname()[0] != 'Windows' and os.environ.get('BITDUST_LOG_USE_COLORS', '1') != '0'
    if exc_value:
        return exception(level, maxTBlevel, exc_info=('', exc_value, []), message=msg)
    return exception(level, maxTBlevel, exc_info, message=msg)


def cb(result, *args, **kwargs):
    _debug = kwargs.pop('debug', 0)
    _debug_level = kwargs.pop('debug_level', 0)
    _method = kwargs.pop('method', 'unknown')
    if _debug and is_debug(_debug_level):
        dbg(_debug_level, 'Deferred.callback() from "%s" method : args=%r  kwargs=%r' % (
            _method,
            args,
            kwargs,
        ))
    return result


def errback(err, *args, **kwargs):
    _debug = kwargs.pop('debug', 0)
    _debug_level = kwargs.pop('debug_level', 0)
    _method = kwargs.pop('method', 'unknown')
    _ignore = kwargs.pop('ignore', False)
    if _debug and is_debug(_debug_level):
        dbg(_debug_level, 'Deferred.errback() from "%s" method with %r : args=%r  kwargs=%r' % (
            _method,
            repr(err).replace('\n', ''),
            args,
            kwargs,
        ))
    if _ignore:
        return None
    return err


def exception(level, maxTBlevel, exc_info, message=None):
    """
    Prints detailed info about last/given exception to STDOUT and also stores exception in a separate file.
    """
    global _LogFileName
    global _StoreExceptionsEnabled
    global _UseColors
    if _UseColors is None:
        _UseColors = platform.uname()[0] != 'Windows' and os.environ.get('BITDUST_LOG_USE_COLORS', '1') != '0'
    if message:
        m = message
        if _UseColors:
            m = '\033[1;31m%s\033[0m' % m
        out(level, m, showtime=True)
    if exc_info is None:
        exc_type, value, trbk = sys.exc_info()
    else:
        exc_type, value, trbk = exc_info
    excArgs = ''
    if value:
        try:
            excArgs = value.__dict__['args']
        except KeyError:
            excArgs = ''
    if trbk:
        excTb = traceback.format_tb(trbk, maxTBlevel)
    else:
        excTb = []
    exc_name = exception_name(value, exc_type, excTb)
    s = 'Exception: <' + exc_name + '>'
    if _UseColors:
        out(level, '\033[1;31m%s\033[0m' % (s.strip()), showtime=True)
    else:
        out(level, s.strip(), showtime=True)
    if excArgs:
        s += '  args:' + excArgs + '\n'
        if _UseColors:
            out(level, '\033[1;31m  args: %s\033[0m' % excArgs)
        else:
            out(level, '  args: %s' % excArgs)
    s += '\n'
    # excTb.reverse()
    for l in excTb:
        s += l + '\n'
        if _UseColors:
            out(level, '\033[1;31m%s\033[0m' % (l.replace('\n', '')))
        else:
            out(level, l.replace('\n', ''))
    if trbk:
        try:
            f_locals = str(repr(trbk.tb_next.tb_next.tb_frame.f_locals))
        except:
            f_locals = ''
        if f_locals:
            out(level, 'locals: %s' % f_locals[:1000])
            s += '\nlocals: %s\n' % f_locals
    if _StoreExceptionsEnabled and _LogFileName:
        if message:
            s = message + '\n' + s
        s = '\n%s\n' % datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f') + s
        exc_label = exc_name.lower().replace(' ', '_').replace('-', '_')[:80]
        exc_label = ''.join([c for c in exc_label if c in '0123456789abcdefghijklmnopqrstuvwxyz_'])
        exc_filename = os.path.join(os.path.dirname(_LogFileName), 'exception_' + exc_label + '.log')
        if os.path.isfile(exc_filename):
            fout = open(exc_filename, 'ab')
        else:
            fout = open(exc_filename, 'wb')
        if sys.version_info[0] == 3:
            if not isinstance(s, bytes):
                s = s.encode('utf-8')
        else:
            if not isinstance(s, str):
                s = s.encode('utf-8')
        s += b'\n==========================================================\n'
        fout.write(s)
        fout.close()
        out(level, 'saved to: %s' % exc_filename)
    return s


def format_exception(maxTBlevel=100, exc_info=None):
    """
    Return string with detailed info about last exception.
    """
    if exc_info is None:
        exc_type, value, trbk = sys.exc_info()
    else:
        exc_type, value, trbk = exc_info
    try:
        excArgs = value.__dict__['args']
    except KeyError:
        excArgs = ''
    excTb = traceback.format_tb(trbk, maxTBlevel)
    tbstring = 'Exception: <' + exception_name(value, exc_type, excTb) + '>\n'
    if excArgs:
        tbstring += '  args:' + excArgs + '\n'
    for s in excTb:
        tbstring += s + '\n'
    return tbstring


def exception_name(value, e_type, tr_back):
    """
    Some tricks to extract the correct exception name from trace-back string.
    """
    try:
        if sys.version_info[0] == 3:
            excStr = str(value)
        else:
            excStr = unicode(value)  # @UndefinedVariable
    except:
        try:
            excStr = repr(value)
        except:
            try:
                excStr = str(value)
            except:
                try:
                    excStr = value.message
                except:
                    excStr = type(value).__name__
    if not excStr:
        if tr_back:
            excStr = tr_back[-1].strip()
    if not excStr:
        excStr = str(e_type)
    return excStr


def set_debug_level(level):
    """
    We will use ``level`` 2-6 for most important things and 10 for really
    minor stuff.

    Level 14 and higher is for things we don't think we want to see
    again. Can set ``level`` to 0 for no debug messages at all.
    """
    global _GlobalDebugLevel
    _GlobalDebugLevel = int(level)


def get_debug_level():
    """
    Returns currently set debug level.
    """
    global _GlobalDebugLevel
    return _GlobalDebugLevel


def get_loging_level(level, return_name=False):
    """
    Find corresponding logging level related to BitDust log level:

        0 : CRITICAL
        1-2 : FATAL
        3-4 : ERROR
        5-6 : WARNING
        7-8 : INFO
        9-10 : DEBUG
        11... : NOTSET
    """
    if level == 0:
        if return_name:
            return 'CRITICAL'
        return logging.CRITICAL
    if level <= 2:
        if return_name:
            return 'FATAL'
        return logging.FATAL
    if level <= 4:
        if return_name:
            return 'ERROR'
        return logging.ERROR
    if level <= 6:
        if return_name:
            return 'WARNING'
        return logging.WARNING
    if level <= 8:
        if return_name:
            return 'INFO'
        return logging.INFO
    if level <= 10:
        if return_name:
            return 'DEBUG'
        return logging.DEBUG
    if return_name:
        return 'NOTSET'
    return logging.NOTSET


def life_begins():
    """
    Start counting time in the logs from that moment.

    If not called the logs will contain current system time.
    """
    global _LifeBeginsTime
    _LifeBeginsTime = time.time()


def when_life_begins():
    global _LifeBeginsTime
    return _LifeBeginsTime


def is_debug(level):
    """
    Return True if something at this ``level`` should be reported given current
    _GlobalDebugLevel.
    """
    global _GlobalDebugLevel
    return _GlobalDebugLevel >= level


def out_globals(level, glob_dict):
    """
    Print all items from dictionary ``glob_dict`` to the logs if current
    _GlobalDebugLevel is higher than ``level``.
    """
    global _GlobalDebugLevel
    if level > _GlobalDebugLevel:
        return
    keys = sorted(glob_dict.keys())
    for k in keys:
        if k != '__builtins__':
            out(level, '%s : %s' % (k, glob_dict[k]))


def time_push(t):
    """
    Remember current system time and set ``t`` marker to that.

    Useful to count execution time of some parts of the code.
    """
    global _TimeTotalDict
    global _TimeDeltaDict
    global _TimeCountsDict
    tm = time.time()
    if t not in _TimeTotalDict:
        _TimeTotalDict[t] = 0.0
        _TimeCountsDict[t] = 0
    _TimeDeltaDict[t] = tm


def time_pop(t):
    """
    Count execution time for marker ``t``.
    """
    global _TimeTotalDict
    global _TimeDeltaDict
    global _TimeCountsDict
    tm = time.time()
    if t not in _TimeTotalDict:
        return
    dt = tm - _TimeDeltaDict[t]
    _TimeTotalDict[t] += dt
    _TimeCountsDict[t] += 1


def print_total_time():
    """
    Print total stats for all time markers.
    """
    global _TimeTotalDict
    global _TimeDeltaDict
    global _TimeCountsDict
    for t in _TimeTotalDict.keys():
        total = _TimeTotalDict[t]
        counts = _TimeCountsDict[t]
        out(2, 'total=%f sec. count=%d, avarage=%f: %s' % (total, counts, total/counts, t))


def exception_hook(typ, value, traceback):
    """
    Callback function to print last exception.
    """
    out(0, 'uncaught exception:')
    exc(exc_info=(typ, value, traceback))


def open_log_file(filename, append_mode=False):
    """
    Open a log file, so all logs will go here instead of STDOUT.
    """
    global _LogFile
    global _LogFileName
    if _LogFile:
        return None
    try:
        if not os.path.isdir(os.path.dirname(os.path.abspath(filename))):
            os.makedirs(os.path.dirname(os.path.abspath(filename)))
        if append_mode:
            _LogFile = open(os.path.abspath(filename), 'a')
        else:
            _LogFile = open(os.path.abspath(filename), 'w')
        _LogFileName = os.path.abspath(filename)
    except:
        _LogFile = None
        _LogFileName = None
    return _LogFile


def close_log_file():
    """
    Closes opened log file.
    """
    global _LogFile
    global _AllLogFiles
    if not _LogFile:
        return
    _LogFile.flush()
    _LogFile.close()
    _LogFile = None
    for logfile in _AllLogFiles.values():
        logfile.flush()
        logfile.close()
    _AllLogFiles.clear()


def open_intercepted_log_file(filename, mode='w'):
    global _InterceptedLogFile
    if not _InterceptedLogFile:
        os.makedirs(os.path.dirname(os.path.abspath(filename)), exist_ok=True)
        _InterceptedLogFile = open(os.path.abspath(filename), mode)
    else:
        warn('intercepted log file %r already opened' % _InterceptedLogFile)


def close_intercepted_log_file():
    global _InterceptedLogFile
    if _InterceptedLogFile:
        _InterceptedLogFile.flush()
        _InterceptedLogFile.close()
        _InterceptedLogFile = None
        return True
    return False


def log_file():
    global _LogFile
    return _LogFile


def log_filename():
    global _LogFileName
    return _LogFileName


def stdout_start_redirecting():
    """
    Replace sys.stdout with STDOUT_redirected so all output get logged via this module.
    """
    global _RedirectStdOut
    global _StdOutPrev
    _RedirectStdOut = True
    _StdOutPrev = sys.stdout
    sys.stdout = STDOUT_redirected()


def stdout_stop_redirecting():
    """
    Restore sys.stdout after ``stdout_start_redirecting``.
    """
    global _RedirectStdOut
    global _StdOutPrev
    _RedirectStdOut = False
    if _StdOutPrev is not None:
        sys.stdout = _StdOutPrev


def stderr_start_redirecting():
    """
    Replace sys.stderr with STDERR_redirected so all errors get logged via this module.
    """
    global _RedirectStdErr
    global _StdErrPrev
    _RedirectStdErr = True
    _StdErrPrev = sys.stderr
    sys.stderr = STDERR_redirected()


def stderr_stop_redirecting():
    """
    Restore sys.stderr after ``stderr_start_redirecting``.
    """
    global _RedirectStdErr
    global _StdErrPrev
    _RedirectStdErr = False
    if _StdErrPrev is not None:
        sys.stderr = _StdErrPrev


def disable_output():
    """
    Disable any output to sys.stdout and sys.stderr
    """
    global _RedirectStdOut
    global _RedirectStdErr
    global _StdOutPrev
    global _StdErrPrev
    global _NoOutput
    _NoOutput = True
    _RedirectStdOut = True
    _RedirectStdErr = True
    _StdOutPrev = sys.stdout
    _StdErrPrev = sys.stderr
    sys.stdout = STDOUT_black_hole()
    sys.stderr = STDERR_black_hole()


def disable_logs():
    """
    Clear _LogsEnabled flag, so calls to ``log()`` and ``exc()`` will do
    nothing.

    Must be used in production release to increase performance. However
    I plan to comment all lines with ``lg.log()`` at all.
    """
    global _LogsEnabled
    _LogsEnabled = False


def logs_enabled():
    global _LogsEnabled
    return _LogsEnabled


def setup_unbuffered_stdout():
    """
    This makes logs to be printed without delays in Linux - unbuffered output.
    Great thanks, the idea is taken from here:
        http://algorithmicallyrandom.blogspot.com/2009/10/python-tips-and-tricks-flushing-stdout.html
    """
    global _OriginalStdOut
    _OriginalStdOut = sys.stdout
    sys.stdout = STDOUT_unbuffered(sys.stdout)


def restore_original_stdout():
    """
    Restore original STDOUT, need to be called after
    ``setup_unbuffered_stdout`` to get back to default state.
    """
    global _OriginalStdOut
    if _OriginalStdOut is None:
        return
    _std_out = sys.stdout
    sys.stdout = _OriginalStdOut
    _OriginalStdOut = None
    try:
        _std_out.close()
    except Exception as exc:
        pass


def setup_unbuffered_stderr():
    """
    This makes error logs to be printed without delays in Linux - unbuffered output.
    """
    global _OriginalStdErr
    _OriginalStdErr = sys.stderr
    sys.stderr = STDERR_unbuffered(sys.stderr)


def restore_original_stderr():
    """
    Restore original STDERR, need to be called after
    ``setup_unbuffered_stderr`` to get back to default state.
    """
    global _OriginalStdErr
    if _OriginalStdErr is None:
        return
    _std_err = sys.stderr
    sys.stderr = _OriginalStdErr
    _OriginalStdErr = None
    try:
        _std_err.close()
    except Exception as exc:
        pass


def set_weblog_func(webstreamfunc):
    """
    Set callback method to be called in Dprint, used to show logs in the WEB
    browser.

    See ``bitdust.lib.weblog`` module.
    """
    global _WebStreamFunc
    _WebStreamFunc = webstreamfunc


#------------------------------------------------------------------------------


class STDOUT_redirected(object):
    """
    Emulate system STDOUT, useful to log any program output.
    """
    softspace = 0

    def read(self):
        pass

    def write(self, s):
        out(0, s.rstrip())

    def flush(self):
        pass

    def close(self):
        pass


class STDERR_redirected(object):
    """
    Emulate system STDERR, useful to log any program error.
    """
    softspace = 0

    def read(self):
        pass

    def write(self, s):
        err(s.rstrip(), level=0)

    def flush(self):
        pass

    def close(self):
        pass


class STDOUT_black_hole(object):
    """
    Useful to disable any output to STDOUT.
    """
    softspace = 0

    def read(self):
        pass

    def write(self, s):
        pass

    def flush(self):
        pass

    def close(self):
        pass


class STDERR_black_hole(object):
    """
    Useful to disable any errors output to STDERR.
    """
    softspace = 0

    def read(self):
        pass

    def write(self, s):
        pass

    def flush(self):
        pass

    def close(self):
        pass


class STDOUT_unbuffered(object):
    def __init__(self, stream):
        self.stream = stream

    def write(self, data):
        try:
            self.stream.write(data)
        except:
            pass
        self.stream.flush()

    def writelines(self, datas):
        try:
            self.stream.writelines(datas)
        except:
            pass
        self.stream.flush()

    def __getattr__(self, attr):
        return getattr(self.stream, attr)


class STDERR_unbuffered(object):
    def __init__(self, stream):
        self.stream = stream

    def write(self, data):
        try:
            self.stream.write(data)
        except:
            pass
        self.stream.flush()

    def writelines(self, datas):
        try:
            self.stream.writelines(datas)
        except:
            pass
        self.stream.flush()

    def __getattr__(self, attr):
        return getattr(self.stream, attr)



#!/usr/bin/python
# p2p_stats.py
#
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (p2p_stats.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: p2p_stats
"""

#------------------------------------------------------------------------------

from bitdust.lib import strng

#------------------------------------------------------------------------------

_PeersProtos = {}
_MyProtos = {}
_CountersIn = {
    'total_bytes': 0,
    'unknown_bytes': 0,
    'total_packets': 0,
    'unknown_packets': 0,
    'failed_packets': 0,
    'identity_cache_count': 0,
    'identity_cache_fails': 0,
    'identity_cache_bytes': 0,
    'peers': {},
}
_CountersOut = {
    'total_bytes': 0,
    'unknown_bytes': 0,
    'total_packets': 0,
    'unknown_packets': 0,
    'failed_packets': 0,
    'peers': {},
}

#------------------------------------------------------------------------------


def my_protos():
    global _MyProtos
    return _MyProtos


def peers_protos():
    global _PeersProtos
    return _PeersProtos


def counters_in():
    global _CountersIn
    return _CountersIn


def counters_out():
    global _CountersOut
    return _CountersOut


#------------------------------------------------------------------------------


def get_total_bytes_in():
    return counters_in()['total_bytes']


def get_total_bytes_out():
    return counters_out()['total_bytes']


#------------------------------------------------------------------------------


def ErasePeerProtosStates(idurl):
    global _PeersProtos
    _PeersProtos.pop(idurl, None)


def EraseAllMyProtosStates():
    my_protos().clear()


def EraseMyProtosStates(idurl):
    my_protos().pop(idurl, None)


#------------------------------------------------------------------------------


def count_outbox(remote_idurl, proto, status, size):
    """

    """
    remote_idurl = strng.to_text(remote_idurl)
    proto = strng.to_text(proto)
    if remote_idurl not in peers_protos():
        peers_protos()[remote_idurl] = set()
    if status == 'finished':
        peers_protos()[remote_idurl].add(proto)

    counters_out()['total_bytes'] += size
    if remote_idurl and remote_idurl.startswith('http://') and remote_idurl.endswith('.xml'):
        if remote_idurl not in counters_out()['peers']:
            counters_out()['peers'][remote_idurl] = 0
        counters_out()['peers'][remote_idurl] += size
        if status == 'finished':
            counters_out()['total_packets'] += 1
        else:
            counters_out()['failed_packets'] += 1
    else:
        counters_out()['unknown_bytes'] += size
        counters_out()['unknown_packets'] += 1


def count_inbox(remote_idurl, proto, status, bytes_received):
    remote_idurl = strng.to_text(remote_idurl)
    proto = strng.to_text(proto)
    if remote_idurl not in my_protos():
        my_protos()[remote_idurl] = set()
    if status == 'finished':
        my_protos()[remote_idurl].add(proto)

    counters_in()['total_bytes'] += bytes_received
    if remote_idurl and remote_idurl.startswith('http://') and remote_idurl.endswith('.xml'):
        if status == 'finished':
            counters_in()['total_packets'] += 1
        else:
            counters_in()['failed_packets'] += 1
        if remote_idurl not in counters_in()['peers']:
            counters_in()['peers'][remote_idurl] = 0
        counters_in()['peers'][remote_idurl] += bytes_received
    else:
        counters_in()['unknown_packets'] += 1
        counters_in()['unknown_bytes'] += bytes_received


def count_identity_cache(idurl, bytes_received):
    if bytes_received > 0:
        counters_in()['total_bytes'] += bytes_received
        counters_in()['identity_cache_count'] += 1
        counters_in()['identity_cache_bytes'] += bytes_received
    else:
        counters_in()['identity_cache_fails'] += 1
